/* postfgr_es_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"



/* Node */

/* 3D solid, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* displacement boundary condition */
int Node_u_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition */
int Node_f_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Node_maxNodes][Node_nDofs];

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];

/* equivalent stress analysis result */
double Node_sigma_eq[Node_maxNodes];



/* Element */

/* 3D solid, linear elastic */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



/* Edge */

#define Edge_maxEdges 100000

int Edge_nEdges;
double Edge_XVec0[Edge_maxEdges][3];
double Edge_XVec1[Edge_maxEdges][3];



/* FaceGroup */

#define N_ELEMENT_NODES Element_nNodes
#if (Element_nNodes == 4)
#define N_FACE_NODES 3
#endif
#if (Element_nNodes == 10)
#define N_FACE_NODES 6
#endif

#define MAX_GROUPS 10000
#define MAX_FACES 1000000

static int NGroups;
static int NTotalFaces;
static int NFaces[MAX_GROUPS];
static int StartFaceId_s[MAX_GROUPS];
static int FaceElementId_s[MAX_FACES];
static int FaceFaceId_s[MAX_FACES];
static int FaceNodeId_s[MAX_FACES][N_FACE_NODES];



void ReadMesh (FILE* fp)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", 
	  &Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1], 
	    &Node_XVec[iNode][2]);
  }
}

void ReadCondition (FILE* fp)
{
  int iNode;
  int iDof_Node;
  int count;
  int iBc;
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
      Node_f_BCVecFlag[iNode][iDof_Node] = 0;
      Node_f_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " displacement BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_u_BCVecFlag[iNode][iDof_Node] = 1;
    Node_u_BCVec[iNode][iDof_Node] = value;
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " load BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_f_BCVecFlag[iNode][iDof_Node] = 1;
    Node_f_BCVec[iNode][iDof_Node] = value;
  }
}

void ReadResult (FILE* fp)
{
  int iNode;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_uVec[iNode][0], 
	    &Node_uVec[iNode][1],
	    &Node_uVec[iNode][2]);
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf", 
	    &Node_sigma_eq[iNode]);
  }
}

void ReadEdge (FILE* fp)
{
  int iEdge;
  
  fscanf (fp, "%d", 
	  &Edge_nEdges);
  assert(Edge_nEdges <= Edge_maxEdges);
  fprintf (stderr, " nEdges %d \n", Edge_nEdges);

  for (iEdge = 0; iEdge < Edge_nEdges; iEdge++) {
    fscanf (fp, "%lf %lf %lf %lf %lf %lf", 
	    &Edge_XVec0[iEdge][0],
	    &Edge_XVec0[iEdge][1], 
	    &Edge_XVec0[iEdge][2],
	    &Edge_XVec1[iEdge][0],
	    &Edge_XVec1[iEdge][1], 
	    &Edge_XVec1[iEdge][2]);
  }
}

void ReadFgr (FILE* fp)
{
  int elementCode;
  int iGroup;
  int iFace;
  int iNode;

  fscanf (fp, "%d", &elementCode);
  assert(elementCode == N_ELEMENT_NODES);
  
  fscanf (fp, "%d", &NGroups);
  assert(NGroups <= MAX_GROUPS);

  NTotalFaces = 0;
  for (iGroup = 0; iGroup < NGroups; iGroup++) {
    fscanf (fp, "%d", &NFaces[iGroup]);
    assert(NTotalFaces + NFaces[iGroup] <= MAX_FACES);
    StartFaceId_s[iGroup] = NTotalFaces;
    NTotalFaces += NFaces[iGroup];

    for (iFace = StartFaceId_s[iGroup]; 
	 iFace < StartFaceId_s[iGroup] + NFaces[iGroup]; 
	 iFace++) {
      fscanf (fp, "%d", &FaceElementId_s[iFace]);
      fscanf (fp, "%d", &FaceFaceId_s[iFace]);
      
      for (iNode = 0; iNode < N_FACE_NODES; iNode++) {
	fscanf (fp, "%d", &FaceNodeId_s[iFace][iNode]); 
      }
    }
  }
}

void ReadData (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 5) {
    fprintf (stderr, 
	     " postfgr_es_3d_tetra input_file output_file edge_file fgr_file\n");
    exit (1);
  }
  
  fp = fopen (argv[1], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open input file  \n");
    exit (1);
  }
  ReadMesh (fp);
  ReadCondition (fp);
  fclose (fp);

  fp = fopen (argv[2], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }
  ReadResult (fp);
  fclose (fp);

  fp = fopen (argv[3], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open edge file  \n");
    exit (1);
  }
  ReadEdge (fp);
  fclose (fp);

  fp = fopen (argv[4], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open fgr file  \n");
    exit (1);
  }
  ReadFgr (fp);
  fclose (fp);
}



double MinRange = 0.0;
double MaxRange = 1.0;

double Scale = 2.0;

double A = 0.0;
double B = 0.0;
double C = 1.0;
double D = 0.0;
double SectionDelta = 0.001;

double LevelValue = 0.0;
double LevelWidth = 1.0;
double LevelDelta = 0.001;

int EdgeIsShown = 1;
int SurfaceIsShown;
int MeshIsShown;
int DeformationIsShown;
int SurfaceContourIsShown;
int SectionContourIsShown;
int IsosurfaceIsShown;
int DisplacementBcIsShown;
int LoadBcIsShown;



static void SetRange ()
{
  int iNode;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double size;
  double midX, midY, midZ;
  double maxLength;

  AutoGL_AssignVector3D 
    (&minX, &minY, &minZ,
     Node_XVec[0][0], Node_XVec[0][1], Node_XVec[0][2]);
  AutoGL_AssignVector3D 
    (&maxX, &maxY, &maxZ,
     Node_XVec[0][0], Node_XVec[0][1], Node_XVec[0][2]);
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double x = Node_XVec[iNode][0];
    double y = Node_XVec[iNode][1];
    double z = Node_XVec[iNode][2];

    AutoGL_GetMinVector3D (&minX, &minY, &minZ,
			   minX, minY, minZ,
			   x, y, z);
    AutoGL_GetMaxVector3D (&maxX, &maxY, &maxZ,
			   maxX, maxY, maxZ,
			   x, y, z);
  }

  size = AutoGL_GetVectorDistance3D 
    (minX, minY, minZ,
     maxX, maxY, maxZ);
  AutoGL_AssignVector3D 
    (&midX, &midY, &midZ,
     (minX + maxX) / 2, 
     (minY + maxY) / 2,
     (minZ + maxZ) / 2);

  AutoGL_SetViewSize (size / 2); 
  AutoGL_SetViewCenter (midX, midY, midZ); 

  SectionDelta = (maxX - minX + maxY - minY + maxZ - minZ) / 100;

  MinRange = Node_sigma_eq[0]; 
  MaxRange = Node_sigma_eq[0]; 
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_sigma_eq[iNode] < MinRange) {
      MinRange = Node_sigma_eq[iNode];
    }
    if (MaxRange < Node_sigma_eq[iNode]) {
      MaxRange = Node_sigma_eq[iNode];
    }
  }

  LevelValue = (MinRange + MaxRange) / 2;
  LevelWidth = (MaxRange - MinRange) / 4;
  LevelDelta = (MaxRange - MinRange) / 100;

  maxLength = 0.0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double ux = Node_uVec[iNode][0];
    double uy = Node_uVec[iNode][1];
    double uz = Node_uVec[iNode][2]; 
    double length = AutoGL_GetVectorLength3D (ux, uy, uz);
    if (maxLength < length) {
      maxLength = length;
    }
  }
    
  Scale = size / maxLength * 0.1;
}

void PlotEdge (void) 
{
  int iEdge;

  AutoGL_SetColor (0, 0, 1);
  for (iEdge = 0; iEdge < Edge_nEdges; iEdge++) {
    AutoGL_DrawLine (Edge_XVec0[iEdge][0],
		     Edge_XVec0[iEdge][1], 
		     Edge_XVec0[iEdge][2],
		     Edge_XVec1[iEdge][0],
		     Edge_XVec1[iEdge][1], 
		     Edge_XVec1[iEdge][2]);
  }
}

void PlotSurface ()
{
  int iFace;

  for (iFace = 0; iFace < NTotalFaces; iFace++) {
    int nodeId0 = FaceNodeId_s[iFace][0];
    int nodeId1 = FaceNodeId_s[iFace][1];
    int nodeId2 = FaceNodeId_s[iFace][2];

    double p0X = Node_XVec[nodeId0][0];
    double p0Y = Node_XVec[nodeId0][1];
    double p0Z = Node_XVec[nodeId0][2];

    double p1X = Node_XVec[nodeId1][0];
    double p1Y = Node_XVec[nodeId1][1];
    double p1Z = Node_XVec[nodeId1][2];

    double p2X = Node_XVec[nodeId2][0];
    double p2Y = Node_XVec[nodeId2][1];
    double p2Z = Node_XVec[nodeId2][2];

    if (DeformationIsShown) {
      p0X += Node_uVec[nodeId0][0] * Scale;
      p0Y += Node_uVec[nodeId0][1] * Scale;
      p0Z += Node_uVec[nodeId0][2] * Scale;
      
      p1X += Node_uVec[nodeId1][0] * Scale;
      p1Y += Node_uVec[nodeId1][1] * Scale;
      p1Z += Node_uVec[nodeId1][2] * Scale;
      
      p2X += Node_uVec[nodeId2][0] * Scale;
      p2Y += Node_uVec[nodeId2][1] * Scale;
      p2Z += Node_uVec[nodeId2][2] * Scale;
    }

    if (MeshIsShown) {
      AutoGL_SetColor (1, 0, 0);
      AutoGL_DrawLine (p0X, p0Y, p0Z,
		       p1X, p1Y, p1Z);
      AutoGL_DrawLine (p1X, p1Y, p1Z,
		       p2X, p2Y, p2Z);
      AutoGL_DrawLine (p2X, p2Y, p2Z,
		       p0X, p0Y, p0Z);
    }

    if (SurfaceContourIsShown) {
      double value0 = (Node_sigma_eq[nodeId0] - MinRange)
	/ (MaxRange - MinRange);
      double value1 = (Node_sigma_eq[nodeId1] - MinRange)
	/ (MaxRange - MinRange);
      double value2 = (Node_sigma_eq[nodeId2] - MinRange)
	/ (MaxRange - MinRange);
      
      AutoGL_DrawContourTriangle (p0X, p0Y, p0Z, value0, 
				  p1X, p1Y, p1Z, value1, 
				  p2X, p2Y, p2Z, value2);
    } else {
      AutoGL_SetColor (0, 1, 1);
      AutoGL_DrawTriangle (p0X, p0Y, p0Z,
			   p1X, p1Y, p1Z,
			   p2X, p2Y, p2Z);
    }
  }
}

void PlotSection (void) 
{
  int iElement;
  
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int iVertex;
    double tetraPoints[4][3];
    double tetraValues[4];
    double tetraGrades[4];
    
    for (iVertex = 0; iVertex < 4; iVertex++) {
      int nodeId = Element_nodeIds[iElement][iVertex];

      tetraPoints[iVertex][0] = Node_XVec[nodeId][0];
      tetraPoints[iVertex][1] = Node_XVec[nodeId][1];
      tetraPoints[iVertex][2] = Node_XVec[nodeId][2];

      tetraValues[iVertex] = Node_sigma_eq[nodeId];
      tetraGrades[iVertex] 
	= (Node_sigma_eq[nodeId] - MinRange) / (MaxRange - MinRange);
    }
    
    AutoGL_DrawTetrahedronSectionContour 
      (A, B, C, D, tetraPoints, tetraGrades);
  }
}

void PlotIsosurfaceAtLevel (double level) 
{
  double grade;
  double red, green, blue;
  int iElement;
  
  grade = (level - MinRange) / (MaxRange - MinRange);
  AutoGL_GetContourColor (&red, &green, &blue,
			  grade);
  AutoGL_SetColor (red, green, blue);
  
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int iVertex;
    double tetraPoints[4][3];
    double tetraValues[4];
    double tetraGrades[4];
    
    for (iVertex = 0; iVertex < 4; iVertex++) {
      int nodeId = Element_nodeIds[iElement][iVertex];

      tetraPoints[iVertex][0] = Node_XVec[nodeId][0];
      tetraPoints[iVertex][1] = Node_XVec[nodeId][1];
      tetraPoints[iVertex][2] = Node_XVec[nodeId][2];

      tetraValues[iVertex] = Node_sigma_eq[nodeId];
      tetraGrades[iVertex] 
	= (Node_sigma_eq[nodeId] - MinRange) / (MaxRange - MinRange);
    }
    
    AutoGL_DrawTetrahedronIsosurface 
      (level, tetraPoints, tetraValues);
  }
}

void PlotIsosurface () 
{
  PlotIsosurfaceAtLevel (LevelValue - LevelWidth);
  PlotIsosurfaceAtLevel (LevelValue);
  PlotIsosurfaceAtLevel (LevelValue + LevelWidth);
}

void PlotDisplacementBc () 
{
  int iNode;

  AutoGL_SetColor (1, 0, 0);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double x = Node_XVec[iNode][0];
    double y = Node_XVec[iNode][1];
    double z = Node_XVec[iNode][2];

    if (Node_u_BCVecFlag[iNode][0]
	&& Node_u_BCVecFlag[iNode][1]
	&& Node_u_BCVecFlag[iNode][2]) {
      
      AutoGL_DrawCubeMark3D (x, y, z, 5, 0);
    } else {
      if (Node_u_BCVecFlag[iNode][0]) {
	double arrowLength = Node_u_BCVec[iNode][0] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x + arrowLength, y, z);
      } 
      if (Node_u_BCVecFlag[iNode][1]) {
	double arrowLength = Node_u_BCVec[iNode][1] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x, y + arrowLength, z);
      } 
      if (Node_u_BCVecFlag[iNode][2]) {
	double arrowLength = Node_u_BCVec[iNode][2] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x, y, z + arrowLength);
      }
    }
  }
}

void PlotLoadBc () 
{
  int iNode;

  AutoGL_SetColor (1, 0, 0);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double x = Node_XVec[iNode][0];
    double y = Node_XVec[iNode][1];
    double z = Node_XVec[iNode][2];

    if (Node_f_BCVecFlag[iNode][0]) {
      double arrowLength = Node_f_BCVec[iNode][0] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x + arrowLength, y, z);
    }
    if (Node_f_BCVecFlag[iNode][1]) {
      double arrowLength = Node_f_BCVec[iNode][1] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x, y + arrowLength, z);
    }
    if (Node_f_BCVecFlag[iNode][2]) {
      double arrowLength = Node_f_BCVec[iNode][2] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x, y, z + arrowLength);
    }
  }
}

void RedrawView (void) 
{
  if (EdgeIsShown) {
    PlotEdge ();
  }
  if (SurfaceIsShown) {
    PlotSurface ();
  }
  if (SectionContourIsShown) {
    PlotSection ();
  }
  if (IsosurfaceIsShown) {
    PlotIsosurface ();
  }
  if (DisplacementBcIsShown) {
    PlotDisplacementBc ();
  }
  if (LoadBcIsShown) {
    PlotLoadBc ();
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}

void UpSection (void) 
{
  D += SectionDelta;
  AutoGL_DrawView ();  
}

void DownSection (void) 
{
  D -= SectionDelta;
  AutoGL_DrawView ();  
}

void UpLevel (void) 
{
  LevelValue += LevelDelta;
  AutoGL_DrawView ();  
}

void DownLevel (void) 
{
  LevelValue -= LevelDelta;
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  ReadData (argc, argv);
  SetRange ();

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode3D (NULL);

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (UpdateModel, "UpdateModel");

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&Scale, "Scale");

  AutoGL_AddBoolean (&EdgeIsShown, "EdgeIsShown");
  AutoGL_SetLabel ("edge");

  AutoGL_AddBoolean (&SurfaceIsShown, "SurfaceIsShown");
  AutoGL_SetLabel ("surface");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_SetLabel ("mesh");
  AutoGL_AddBoolean (&DeformationIsShown, "DeformationIsShown");
  AutoGL_SetLabel ("deformation");
  AutoGL_AddBoolean (&SurfaceContourIsShown, "SurfaceContourIsShown");
  AutoGL_SetLabel ("surface contour");

  AutoGL_AddBoolean (&DisplacementBcIsShown, "DisplacementBcIsShown");
  AutoGL_SetLabel ("disp. bc");
  AutoGL_AddBoolean (&LoadBcIsShown, "LoadBcIsShown");
  AutoGL_SetLabel ("load bc");

  AutoGL_AddBoolean (&SectionContourIsShown, "SectionContourIsShown");
  AutoGL_SetLabel ("section contour");
  AutoGL_SetLabel ("equation Ax+By+Cz+D=0");
  AutoGL_AddReal (&A, "A");      
  AutoGL_AddReal (&B, "B");      
  AutoGL_AddReal (&C, "C");      
  AutoGL_AddReal (&D, "D");      
  AutoGL_AddReal (&SectionDelta, "SectionDelta");
  AutoGL_AddCallback (UpSection, "UpSection");
  AutoGL_AddCallback (DownSection, "DownSection");

  AutoGL_AddBoolean (&IsosurfaceIsShown, "IsosurfaceIsShown");
  AutoGL_SetLabel ("isosurface");
  AutoGL_AddReal (&LevelValue, "LevelValue");
  AutoGL_AddReal (&LevelWidth, "LevelWidth");
  AutoGL_AddReal (&LevelDelta, "LevelDelta");
  AutoGL_AddCallback (UpLevel, "UpLevel");
  AutoGL_AddCallback (DownLevel, "DownLevel");
}






