/* autolinas_ldl_skyline_block.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_ldl_inside.h"
#include "autolinas_ldl_skyline_inside.h"



/* symmetric matrix */
/* direct method, LtDL decomposition */
/* profile form (IJK), block */
/* sparse matrix, skyline storage, lower triangle */



/* module variables */

#if 0
#define BLOCK_SIZE 36
#else
#define BLOCK_SIZE 72
#endif

static int System_nBlocks;
static int System_minBlockIndexs[System_maxDofs / BLOCK_SIZE + 1];

static double DA[BLOCK_SIZE];               /* diagonal array */
static double A0[BLOCK_SIZE];               /* temporary array */
static double SB0[BLOCK_SIZE][BLOCK_SIZE];  /* source block 0 */
static double SB1[BLOCK_SIZE][BLOCK_SIZE];  /* source block 1 */
static double DB[BLOCK_SIZE][BLOCK_SIZE];   /* destination block 0 */



static void CheckBlock (void)
{
  int bi;
  int i;

  System_nBlocks = (System_nDofs + BLOCK_SIZE - 1) / BLOCK_SIZE;
  for (bi = 0; bi < System_nBlocks; bi++) {
    int iMin = System_nDofs;

    for (i = 0; i < BLOCK_SIZE; i++) {
      int I = bi * BLOCK_SIZE + i;

      if (I < System_nDofs) {
	if (System_minIndexs[I] < iMin) {
	  iMin = System_minIndexs[I];
	}
      }
    }
    System_minBlockIndexs[bi] = iMin / BLOCK_SIZE;
  }
}

static void ReadDiagonal 
(int bi,
 double A[BLOCK_SIZE])
{
  int i;

  for (i = 0; i < BLOCK_SIZE; i++) {
    int I = bi * BLOCK_SIZE + i;
    if (I < System_nDofs) {
      A[i] = SM[System_ij[I] + I];
    } else {
      A[i] = 1.0;
    }
  }

#ifdef MEASURE
  AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * 2);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE);
#endif
}

static void ReadBlock 
(int bi, int bj,
 double B[BLOCK_SIZE][BLOCK_SIZE])
{
  int i, j;

  for (i = 0; i < BLOCK_SIZE; i++) {
    int I = bi * BLOCK_SIZE + i;
    if (I < System_nDofs) {
      for (j = 0; j < BLOCK_SIZE; j++) {
	int J = bj * BLOCK_SIZE + j;
	if (System_minIndexs[I] <= J && J <= I) {
	  B[i][j] = SM[System_ij[I] + J];
	} else {
	  B[i][j] = 0.0;
	}
      }
    } else {
      for (j = 0; j < BLOCK_SIZE; j++) {
	B[i][j] = 0.0;
      }
      B[i][i] = 1.0;
    }
  }

#ifdef MEASURE
  if (bi == bj) {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE / 2);
  } else {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE);
  }
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE);
#endif
}

static void WriteBlock 
(int bi, int bj,
 double B[BLOCK_SIZE][BLOCK_SIZE])
{
  int i, j;

  for (i = 0; i < BLOCK_SIZE; i++) {
    int I = bi * BLOCK_SIZE + i;

    if (I < System_nDofs) {

      for (j = 0; j < BLOCK_SIZE; j++) {
	int J = bj * BLOCK_SIZE + j;

	if (System_minIndexs[I] <= J && J <= I) {
	  SM[System_ij[I] + J] = B[i][j];
	}
      }
    }
  }

#ifdef MEASURE
  if (bi == bj) {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE / 2);
  } else {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE);
  }
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE);
#endif
}

/* SB0 (bj, bl), SB0 (bj, bl) -> DB (bj, bj) */
static void DoBlock_SJL_SJL_DJJ (void)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j++) {

    for (l = 0; l < BLOCK_SIZE; l++) {
      A0[l] = DA[l] * SB0[j][l];
    }

    for (i = j; i < BLOCK_SIZE; i++) {

      for (l = 0; l < BLOCK_SIZE; l++) {
	DB[i][j] -= A0[l] * SB0[i][l];
      }
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 2);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 2);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE * 2 / 2);
#endif
}

/* DB (bj, bj), DB (bj, bj) -> DB (bj, bj) */
static void DoBlock_SJJ_SJJ_DJJ (void)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j++) {

    for (l = 0; l < j; l++) {
      DB[j][j] -= DB[l][l] * DB[j][l] * DB[j][l];
    }

    for (i = j + 1; i < BLOCK_SIZE; i++) {

      for (l = 0; l < j; l++) {
	DB[i][j] -= DB[l][l] * DB[j][l] * DB[i][l];
      }

      DB[i][j] /= DB[j][j];
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 4);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE * 2 / 4);
  AutoLinAS_AddOp (M_OTHER_OP, BLOCK_SIZE * BLOCK_SIZE / 2);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE * 3 / 4);
#endif
}

/* SB0 (bj, bl), SB1 (bi, bl) -> DB (bi, bj) */
static void DoBlock_SJL_SIL_DIJ (void)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j++) {

    for (l = 0; l < BLOCK_SIZE; l++) {
      A0[l] = DA[l] * SB0[j][l];
    }

    for (i = 0; i < BLOCK_SIZE; i++) {

      for (l = 0; l < BLOCK_SIZE; l++) {
	DB[i][j] -= A0[l] * SB1[i][l];
      }
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE * 2);
#endif
}

/* SB0 (bj, bj), DB (bi, bj) -> DB (bi, bj) */
static void DoBlock_SJJ_SIJ_DIJ (void)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j++) {
    for (i = 0; i < BLOCK_SIZE; i++) {

      for (l = 0; l < j; l++) {
	DB[i][j] -= SB0[l][l] * SB0[j][l] * DB[i][l];
      }

      DB[i][j] /= SB0[j][j];
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 2);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE * 2 / 2);
  AutoLinAS_AddOp (M_OTHER_OP, BLOCK_SIZE * BLOCK_SIZE);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE * 3 / 2);
#endif
}

void AutoLinAS_INSIDE_LDLDecomposition (void)
{
  int bi, bj, bl;

  CheckBlock ();

  for (bj = 0; bj < System_nBlocks; bj++) {

    ReadBlock (bj, bj, 
	       DB);

    for (bl = System_minBlockIndexs[bj]; bl < bj; bl++) {
      ReadDiagonal (bl, 
		    DA);
      ReadBlock (bj, bl, 
		 SB0);
      DoBlock_SJL_SJL_DJJ (); 
      /* DA (bl), SB0 (bj, bl), SB0 (bj, bl) -> DB (bj, bj) */
    }

    DoBlock_SJJ_SJJ_DJJ (); 
    /* DB (bj, bj), DB (bj, bj) -> DB (bj, bj) */
    WriteBlock (bj, bj, DB);

    for (bi = bj + 1; bi < System_nBlocks; bi++) {
      if (bj < System_minBlockIndexs[bi]) {
	continue;
      }

      ReadBlock (bi, bj, 
		 DB);

      for (bl = System_minBlockIndexs[bj]; bl < bj; bl++) {
	if (bl < System_minBlockIndexs[bi]) {
	  continue;
	}

	ReadDiagonal (bl, 
		      DA);
	ReadBlock (bj, bl, 
		   SB0);
	ReadBlock (bi, bl, 
		   SB1);
	DoBlock_SJL_SIL_DIJ (); 
	/* DA (bl), SB0 (bj, bl), SB1 (bi, bl) -> DB (bi, bj) */
      }

      ReadBlock (bj, bj, 
		 SB0);
      DoBlock_SJJ_SIJ_DIJ (); 
      /* SB0 (bj, bj), DB (bi, bj) -> DB (bi, bj) */
      WriteBlock (bi, bj, DB);
    }
  }
}

void AutoLinAS_INSIDE_ForwardReduction (void)
{
  int i, j;

#if (Node_nDofs == 1)
  for (i = 0; i < System_nDofs; i++) {
    double sum = 0.0;

    for (j = System_minIndexs[i]; j < i; j++) {
      sum += SM[System_ij[i] + j] * SV1[j];
    }
    SV1[i] -= sum;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif

#endif
#if (Node_nDofs == 2)

  for (i = 0; i < System_nDofs; i += 2) {
    double sum0 = 0.0;
    double sum1 = 0.0;

    for (j = System_minIndexs[i]; j < i; j++) {
      sum0 += SM[System_ij[i + 0] + j] * SV1[j];
      sum1 += SM[System_ij[i + 1] + j] * SV1[j];
    }
    SV1[i + 0] -= sum0;

    sum1 += SM[System_ij[i + 1] + i] * SV1[i];
    SV1[i + 1] -= sum1;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 1.5);
#endif

#endif
#if (Node_nDofs == 3)

  for (i = 0; i < System_nDofs; i += 3) {
    double sum0 = 0.0;
    double sum1 = 0.0;
    double sum2 = 0.0;

    for (j = System_minIndexs[i]; j < i; j++) {
      sum0 += SM[System_ij[i + 0] + j] * SV1[j];
      sum1 += SM[System_ij[i + 1] + j] * SV1[j];
      sum2 += SM[System_ij[i + 2] + j] * SV1[j];
    }
    SV1[i + 0] -= sum0;

    sum1 += SM[System_ij[i + 1] + i] * SV1[i];
    sum2 += SM[System_ij[i + 2] + i] * SV1[i];
    SV1[i + 1] -= sum1;

    sum2 += SM[System_ij[i + 2] + i + 1] * SV1[i + 1];
    SV1[i + 2] -= sum2;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 1.333);
#endif

#endif

}

void AutoLinAS_INSIDE_BackwardSubstitution (void)
{
  int i, j;
  int maxBandWidth;

  maxBandWidth = -System_minIndexs[0];
  for (i = 1; i < System_nDofs; i++) {
    if (maxBandWidth < i - System_minIndexs[i]) {
      maxBandWidth = i - System_minIndexs[i];
    }
  }

  SV1[System_nDofs - 1] 
    /= SM[System_ij[System_nDofs - 1] + System_nDofs - 1];

  for (i = System_nDofs - 2; i >= 0; i--) {
    int jMin;

    jMin = i + maxBandWidth;
    if (jMin > System_nDofs - 1) {
      jMin = System_nDofs - 1;
    } else {
      while (System_minIndexs[jMin] > i && jMin != i) {
	jMin--;
      }
    }
    SV1[i] /= SM[System_ij[i] + i];

    for (j = i + 1; j <= jMin; j++) {
      if (i >= System_minIndexs[j]) {
        SV1[i] -= SM[System_ij[j] + i] * SV1[j];
      }
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif
}

