/* main_es_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autolinas_predefined.h>
#include <autolinas.h>

#include "matrix.h"
#include "es_2d_tri.h"



/* System */

/* 2D plane strain, linear elastic, static analysis */

static int ElementNodeIdCallback (int iElement, int iNode_Elem)
{
  return Element_nodeIds[iElement][iNode_Elem];
}

static void Solve (void)
{
  int iNode;
  int iElement;
  int iDof_Node;

  AutoLinAS_Initialize (Node_nNodes, Element_nElements, 
			ElementNodeIdCallback);

  /* _{u_fixed} = {0}; */
  AutoLinAS_ClearConstraint ();

  /* _{u_fixed} = _node._{u_BC}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) {
	double value = Node_u_BCVec[iNode][iDof_Node];

	AutoLinAS_SetConstraint (iNode, iDof_Node, value);
      }
    }
  }

  /* _{f} = {0}; */
  AutoLinAS_ClearRightHandSide ();

  /* _{f} += _element._{f_b}; */
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    double tmp0[Element_nDofs];
    
    Element_f_bVec (iElement,
		    tmp0);
    
    AutoLinAS_AddElementVectorIntoRightHandSide (iElement, tmp0);
  }

  /* _{f} += _node._{f_BC}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) {
	double value = Node_f_BCVec[iNode][iDof_Node];

	AutoLinAS_AddValueIntoRightHandSide (iNode, iDof_Node, value);
      }
    }
  }

  /* _[K] = _element._[K]; */
  /* _[K].fix (_{u_fixed}); */
  AutoLinAS_SetElementMatrixCallback (Element_KMat);
  AutoLinAS_Fix ();

  /* _{u} = _[K].solveSystem (_{f}); */
  AutoLinAS_Solve ();
  
  /* _node._{u} = _{u}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_uVec[iNode][iDof_Node] = AutoLinAS_Result (iNode, iDof_Node);
    }
  }
}

static void ReadMesh (FILE* fp)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf ", 
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1]);
  }
}

static void ReadCondition (FILE* fp)
{
  int iNode;
  int iDof_Node;
  int count;
  int iBc;
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
      Node_f_BCVecFlag[iNode][iDof_Node] = 0;
      Node_f_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " displacement BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_u_BCVecFlag[iNode][iDof_Node] = 1;
    Node_u_BCVec[iNode][iDof_Node] = value;
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " load BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_f_BCVecFlag[iNode][iDof_Node] = 1;
    Node_f_BCVec[iNode][iDof_Node] = value;
  }
}

#define EPS 0.00001

static void MakeCondition (void)
{
  int iNode;
  int iDof_Node;
  double minY, maxY;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  minY = Node_XVec[0][1];
  maxY = Node_XVec[0][1];
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_XVec[iNode][1] < minY) {
      minY = Node_XVec[iNode][1];
    }
    if (maxY < Node_XVec[iNode][1]) {
      maxY = Node_XVec[iNode][1];
    }
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (fabs (Node_XVec[iNode][1] - minY) < EPS) {
      /* bottom edge */
      Node_u_BCVecFlag[iNode][0] = 1; 
      Node_u_BCVecFlag[iNode][1] = 1; 
      Node_u_BCVec[iNode][0] = 0.0;
      Node_u_BCVec[iNode][1] = 0.0;
    } else if (fabs (Node_XVec[iNode][1] - maxY) < EPS) {
      /* top edge */
      Node_u_BCVecFlag[iNode][1] = 1; 
      Node_u_BCVec[iNode][1] = 0.1;
    }
  }

  Material_gVec[0] = 0.0;
  Material_gVec[1] = 0.0;

  Material_E = 1.0;
  Material_nyu = 0.3;
  Material_rou = 1.0;
}

static void PrintResult (FILE* fp)
{
  int iNode;
  int iDof_Node;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      fprintf (fp, "%s ", 
	       RealString (Node_uVec[iNode][iDof_Node], 7, 3));
    }

    fprintf (fp, "\n");
  }
  fprintf (fp, "\n");
}

int main (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 3) {
    fprintf (stderr, 
	     " main_es_2d_tri input_file_name output_file_name \n");
    exit (1);
  }

  fp = fopen (argv[1], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open input file  \n");
    exit (1);
  }
  ReadMesh (fp);
#if 0
  ReadCondition (fp);
#else
  MakeCondition ();
#endif
  fclose (fp);

  Solve ();
  
  fp = fopen (argv[2], "w");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }

  PrintResult (fp);

  fclose (fp);

  return 0;
}


