/* main_es_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autolinas_predefined.h>
#include <autolinas.h>

#include "matrix.h"
#include "es_3d_tetra.h"



/* System */

/* 3D solid, linear elastic, static analysis */

static int ElementNodeIdCallback (int iElement, int iNode_Elem)
{
  return Element_nodeIds[iElement][iNode_Elem];
}

static void Solve (void)
{
  int iNode;
  int iElement;
  int iDof_Node;
  int iNode_Elem;

  AutoLinAS_Initialize (Node_nNodes, Element_nElements, 
			ElementNodeIdCallback);

  /* _{u_fixed} = {0}; */
  AutoLinAS_ClearConstraint ();

  /* _{u_fixed} = _node._{u_BC}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) {
	double value = Node_u_BCVec[iNode][iDof_Node];

	AutoLinAS_SetConstraint (iNode, iDof_Node, value);
      }
    }
  }

  /* _{f} = {0}; */
  AutoLinAS_ClearRightHandSide ();

  /* _{f} += _element._{f_b}; */
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    double tmp0[Element_nDofs];
    
    Element_f_bVec (iElement,
		    tmp0);
    
    AutoLinAS_AddElementVectorIntoRightHandSide (iElement, tmp0);
  }

  /* _{f} += _node._{f_BC}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) {
	double value = Node_f_BCVec[iNode][iDof_Node];

	AutoLinAS_AddValueIntoRightHandSide (iNode, iDof_Node, value);
      }
    }
  }

  /* _[K] = _element._[K]; */
  /* _[K].fix (_{u_fixed}); */
  AutoLinAS_SetElementMatrixCallback (Element_KMat);
  AutoLinAS_Fix ();

  /* _{u} = _[K].solveSystem (_{f}); */
  AutoLinAS_Solve ();
  
  /* _node._{u} = _{u}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_uVec[iNode][iDof_Node] = AutoLinAS_Result (iNode, iDof_Node);
    }
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    Node_weight[iNode] = 0.0;
  }
  
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int iNode = Element_nodeIds[iElement][iNode_Elem];
      
      Node_weight[iNode] += 1.0;
    }
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    Node_sigma_eq[iNode] = 0.0;
  }

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    Element_calculate_sigma_eq (iElement);    
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    Node_sigma_eq[iNode] /= Node_weight[iNode];
  }
}

static void ReadMesh (FILE* fp)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf %lf ", 
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1],
	    &Node_XVec[iNode][2]);
  }
}

static void ReadCondition (FILE* fp)
{
  int iNode;
  int iDof_Node;
  int count;
  int iBc;
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
      Node_f_BCVecFlag[iNode][iDof_Node] = 0;
      Node_f_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " displacement BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_u_BCVecFlag[iNode][iDof_Node] = 1;
    Node_u_BCVec[iNode][iDof_Node] = value;
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " load BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_f_BCVecFlag[iNode][iDof_Node] = 1;
    Node_f_BCVec[iNode][iDof_Node] = value;
  }

  Material_gVec[0] = 0.0;
  Material_gVec[1] = 0.0;
  Material_gVec[2] = 0.0;

  Material_E = 1.0;
  Material_nyu = 0.3;
  Material_rou = 1.0;
}

#define EPS 0.00001

static void MakeCondition (void)
{
  int iNode;
  int iDof_Node;
  double minZ, maxZ;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  minZ = Node_XVec[0][2];
  maxZ = Node_XVec[0][2];
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_XVec[iNode][2] < minZ) {
      minZ = Node_XVec[iNode][2];
    }
    if (maxZ < Node_XVec[iNode][2]) {
      maxZ = Node_XVec[iNode][2];
    }
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (fabs (Node_XVec[iNode][2] - minZ) < EPS) {
      /* bottom face */
      Node_u_BCVecFlag[iNode][0] = 1; 
      Node_u_BCVecFlag[iNode][1] = 1; 
      Node_u_BCVecFlag[iNode][2] = 1; 
      Node_u_BCVec[iNode][0] = 0.0;
      Node_u_BCVec[iNode][1] = 0.0;
      Node_u_BCVec[iNode][2] = 0.0;
    } else if (fabs (Node_XVec[iNode][2] - maxZ) < EPS) {
#if 1
      /* top face */
      Node_u_BCVecFlag[iNode][2] = 1; 
      Node_u_BCVec[iNode][2] = 0.1;
#endif
    }
  }

#if 0
  Material_gVec[0] = 0.0;
  Material_gVec[1] = 0.0;
  Material_gVec[2] = 0.0;
#endif
#if 0
  Material_gVec[0] = 1.0;
  Material_gVec[1] = 1.0;
  Material_gVec[2] = 0.0;
#endif
#if 0
  Material_gVec[0] = 0.0;
  Material_gVec[1] = 0.0;
  Material_gVec[2] = 1.0;
#endif

  Material_E = 1.0;
  Material_nyu = 0.3;
  Material_rou = 1.0;
}

static void PrintResult (FILE* fp)
{
  int iNode;
  int iDof_Node;

  /* displacement analysis result */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

#if 1
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      fprintf (fp, "%s ", 
	       RealString (Node_uVec[iNode][iDof_Node], 7, 3));
    }
#else
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      fprintf (fp, "%g ", Node_uVec[iNode][iDof_Node]);
    }
#endif

    fprintf (fp, "\n");
  }
  fprintf (fp, "\n");

  /* nodal equivalent stress analysis result */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    
#if 1
    fprintf (fp, "%s ", 
	     RealString (Node_sigma_eq[iNode], 7, 3));
#else
    fprintf (fp, "%g ", Node_sigma_eq[iNode]);
#endif

    fprintf (fp, "\n");
  }
  fprintf (fp, "\n");
}

/* for AutoMESH */
static void PrintAutoMESHResult ()
{
  FILE* fp;
  int iNode;
  int iDof_Node;

  fp = fopen ("displacement.vector", "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Node_nNodes);

  /* displacement analysis result */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%d ", 
	     iNode);
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      fprintf (fp, "%lf ", 
	       Node_uVec[iNode][iDof_Node]);
    }
    fprintf (fp, "\n");
  }

  fclose (fp);



  fp = fopen ("stress.scalar", "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Node_nNodes);

  /* nodal equivalent stress analysis result */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%d %lf\n", 
	     iNode, 
	     Node_sigma_eq[iNode]);
  }

  fclose (fp);
}

int main (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 3) {
    fprintf (stderr, 
	     " main_es_3d_tetra input_file_name output_file_name \n");
    exit (1);
  }

  fp = fopen (argv[1], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open input file  \n");
    exit (1);
  }
  ReadMesh (fp);
#if 0
  ReadCondition (fp);
#else
  MakeCondition ();
#endif
  fclose (fp);

  Solve ();

#if 0  
  fp = fopen (argv[2], "w");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }

  PrintResult (fp);

  fclose (fp);
#else
  PrintAutoMESHResult ();
#endif

  return 0;
}


