/* make_input_file.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <Adv/AdvDocument.h>

#include "../lib/autolinas_predefined.h"



#define Face_maxFaces 1000000



int Domain_nDomains;
int Node_nGlobalNodes;

int Node_nNodes;
int Element_nElements;

/* global coordinate */
/* part node ID --> coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* part node ID --> global node ID */
int Node_partToGlobalIds[Node_maxNodes];

/* subdomain node ID --> part node ID */
int Node_subdomainToPartIds[Node_maxNodes];

/* connectivity (in the subdomain) */
/* element ID --> part node ID */
int Element_nodeIds[Element_maxElements][Element_nNodes];

/* displacement boundary condition (in the subdomain) */
/* part node ID --> flag & value */
int Node_u_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition (in the subdomain) */
/* part node ID --> flag & value */
int Node_f_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Node_maxNodes][Node_nDofs];

/* subdomain (part) face ID --> subdomain (part) element ID */
int Face_nFaces;
int Face_elementId[Face_maxFaces];
int Face_faceId[Face_maxFaces];



void ReadNode (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nDimensions;

  doc = adv_dbox_find_by_property 
    (dbox, NULL, "content_type", "Node", NULL);
  assert(doc != NULL);

  adv_dio_get_property_int32 (doc, "num_items", 
			      &Node_nNodes);
  assert(0 < Node_nNodes);
  assert(Node_nNodes <= Node_maxNodes);

  adv_dio_get_property_int32 (doc, "num_items_orig", 
			      &Node_nGlobalNodes);
  assert(0 < Node_nGlobalNodes);
  assert(Node_nNodes <= Node_nGlobalNodes);

  fprintf (stderr, " Node_nNodes %d \n", Node_nNodes);

  adv_dio_get_property_int32 (doc, "dimension", &nDimensions);
  assert(nDimensions == 3);

  offset = 0;

  /* ordered by part node ID */
  offset += adv_dio_read_float64v (doc, offset, Node_nNodes * nDimensions, 
				   (void*)Node_XVec);

  adv_dio_close (doc);
}

void ReadPartToGlobal (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nItems;

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "FEGenericAttribute", 
     "label", "NodeIndex_PartToGlobal", NULL);
  assert(doc != NULL);

  adv_dio_get_property_int32 (doc, "num_items", 
			      &nItems);
  assert(nItems == Node_nNodes);

  offset = 0;

  offset += adv_dio_read_int32v (doc, offset, Node_nNodes, 
				 (void*)Node_partToGlobalIds);

  adv_dio_close (doc);
}

void ReadSubdomainToPart (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nDomains;
  int nItems;

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "HDDM_FEGenericAttribute", 
     "label", "NodeIndex_SubdomainToPart", NULL);
  assert(doc != NULL);

  adv_dio_get_property_int32 (doc, "num_subdomains", 
			      &nDomains);
  assert(nDomains == 1);

  offset = 0;

  offset += adv_dio_read_int32 (doc, offset, 
				&nItems);
  assert(nItems == Node_nNodes);

  offset += adv_dio_read_int32v (doc, offset, Node_nNodes, 
				 (void*)Node_subdomainToPartIds);

  adv_dio_close (doc);
}

void ReadElement (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  char elementTypeName[256];
  int nDomains;
  int nNodesPerElement;
  int iElement;
  int iNode_Elem;

  doc = adv_dbox_find_by_property 
    (dbox, NULL, "content_type", "HDDM_Element", NULL);
  assert(doc != NULL);

  strcpy (elementTypeName, adv_dio_get_property (doc, "element_type"));
  if (strcmp (elementTypeName, "3DLinearTetrahedron") == 0) {
    assert(Element_nNodes == 4);
  } else if (strcmp (elementTypeName, "3DQuadraticTetrahedron") == 0) {
    assert(Element_nNodes == 10);
  } else {
    assert(0);
  }

  adv_dio_get_property_int32 (doc, "num_subdomains", 
			      &nDomains);
  assert(nDomains == 1);

  adv_dio_get_property_int32 (doc, "num_nodes_per_element", 
			      &nNodesPerElement);
  assert(nNodesPerElement == Element_nNodes);

  offset = 0;

  offset += adv_dio_read_int32 (doc, offset, 
				&Element_nElements);
  assert(0 < Element_nElements);
  assert(Element_nElements <= Element_maxElements);

  fprintf (stderr, " Element_nElements %d \n", Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int subdomainNodeId;
      int partNodeId;
      
      offset += adv_dio_read_int32 (doc, offset, 
				    &subdomainNodeId);
      partNodeId = Node_subdomainToPartIds[subdomainNodeId];
      Element_nodeIds[iElement][iNode_Elem] = partNodeId;
    }
  }
      
  adv_dio_close (doc);
}

void ReadDisplacementBc (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nDomains;
  int nItems;
  int iItem;
  int iNode;
  int iDof_Node;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "HDDM_FEGenericAttribute", 
     "label", "ForcedDisplacement", NULL);
  if (doc == NULL) {
    fprintf (stderr, " no ForcedDisplacement \n");
    return;
  }

  adv_dio_get_property_int32 (doc, "num_subdomains", 
			      &nDomains);
  assert(nDomains == 1);

  offset = 0;

  offset += adv_dio_read_int32 (doc, offset, 
				&nItems);
  assert(0 <= nItems);

  for (iItem = 0; iItem < nItems; iItem++) {
    int subdomainNodeId;
    int partNodeId;
    int dofId;
    double value;

    offset += adv_dio_read_int32 (doc, offset, 
				  &subdomainNodeId);
    partNodeId = Node_subdomainToPartIds[subdomainNodeId];
    offset += adv_dio_read_int32 (doc, offset, 
				  &dofId);
    offset += adv_dio_read_float64 (doc, offset, 
				    &value);

    Node_u_BCVecFlag[partNodeId][dofId] = 1;
    Node_u_BCVec[partNodeId][dofId] = value;
  }

  adv_dio_close (doc);
}

void ReadLoadBc (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nDomains;
  int nItems;
  int iItem;
  int iNode;
  int iDof_Node;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_f_BCVecFlag[iNode][iDof_Node] = 0;
      Node_f_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "HDDM_FEGenericAttribute", 
     "label", "Load", NULL);
  if (doc == NULL) {
    fprintf (stderr, " no Load \n");
    return;
  }

  adv_dio_get_property_int32 (doc, "num_subdomains", 
			      &nDomains);
  assert(nDomains == 1);

  offset = 0;

  offset += adv_dio_read_int32 (doc, offset, 
				&nItems);
  assert(0 <= nItems);

  for (iItem = 0; iItem < nItems; iItem++) {
    int subdomainNodeId;
    int partNodeId;
    int dofId;
    double value;

    offset += adv_dio_read_int32 (doc, offset, 
				  &subdomainNodeId);
    partNodeId = Node_subdomainToPartIds[subdomainNodeId];
    offset += adv_dio_read_int32 (doc, offset, 
				  &dofId);
    offset += adv_dio_read_float64 (doc, offset, 
				    &value);

    Node_f_BCVecFlag[partNodeId][dofId] = 1;
    Node_f_BCVec[partNodeId][dofId] = value;
  }

  adv_dio_close (doc);
}

void ReadFaceGroup (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nDomains;
  int iFace;

  Face_nFaces = 0;

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "HDDM_FEGenericAttribute", 
     "label", "FaceGroup", NULL);
  if (doc == NULL) {
    fprintf (stderr, " no FaceGroup \n");
    return;
  }

  adv_dio_get_property_int32 (doc, "num_subdomains", 
			      &nDomains);
  assert(nDomains == 1);
  
  offset = 0;
  
  offset += adv_dio_read_int32 (doc, offset, 
				&Face_nFaces);
  assert(0 <= Face_nFaces);
  assert(Face_nFaces < Face_maxFaces);

  /* subdomain face ID is identical to part face ID */
  for (iFace = 0; iFace < Face_nFaces; iFace++) {
    int subdomainElementId;
    int faceId;

    offset += adv_dio_read_int32 (doc, offset, 
				  &subdomainElementId);
    offset += adv_dio_read_int32 (doc, offset, 
				  &faceId);

    /* subdomain element ID is identical to part element ID */
    Face_elementId[iFace] = subdomainElementId;
    Face_faceId[iFace] = faceId;
  }

  adv_dio_close (doc);
}

void ReadAdvIOFile (char* fileName)
{
  AdvDatabox* dbox;

  dbox = adv_dbox_new ();

  if (!adv_dbox_add (dbox, fileName)) {
    fprintf (stderr, "Cannot open %s\n", fileName);
    exit (1);
  }

  ReadNode (dbox);
  ReadPartToGlobal (dbox);
  ReadSubdomainToPart (dbox);

  /* all the node IDs are subdomain node ID, so conversion from subdomain 
     node ID to part node ID is required */
  ReadElement (dbox);
  ReadDisplacementBc (dbox);
  ReadLoadBc (dbox);
  ReadFaceGroup (dbox);

  adv_dbox_close (dbox);
}

void WriteDomainFile (char* fileName)
{
  FILE* fp;
  int iNode;
  
  fp = fopen (fileName, "w");
  assert(fp);
  
  fprintf (fp, "%d\n", Node_nNodes);

  /* ordered by part node ID */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int globalNodeId = Node_partToGlobalIds[iNode];

    fprintf (fp, "%d\n", globalNodeId);
  }
  
  fclose (fp);
}

void CheckMesh ()
{
  static int nodeFlag[Node_maxNodes];
  int iNode;
  int iElement;
  int iNode_Elem;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    nodeFlag[iNode] = 0;
  }

  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int nodeId = Element_nodeIds[iElement][iNode_Elem];

      assert(0 <= nodeId);
      assert(nodeId < Node_nNodes);
      nodeFlag[nodeId] = 1;
    }
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    assert(nodeFlag[iNode]);
  }
}

void WriteInputFile (char* fileName)
{
  FILE* fp;
  int iElement;
  int iNode_Elem;
  int iNode;
  int iDof_Node;
  int count;
      
  fp = fopen (fileName, "w");
  assert(fp);

  /* element connectivity */

  fprintf (fp, "%d\n", Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iNode_Elem]);
    }

    fprintf (fp, "\n");
  }
  
  /* nodal coordinate */

  fprintf (fp, "%d\n", Node_nNodes);

  /* ordered by part node ID */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%f %f %f \n", 
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1],
	     Node_XVec[iNode][2]);
  }
  
  /* displacement B.C. */

  count = 0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) count++;
    }
  }

  fprintf (fp, "%d\n", count);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) {
	fprintf (fp, "%d %d %f \n", 
		 iNode,
		 iDof_Node,
		 Node_u_BCVec[iNode][iDof_Node]);
      }
    }
  }

  /* load B.C. */

  count = 0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) count++;
    }
  }

  fprintf (fp, "%d\n", count);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) {
	fprintf (fp, "%d %d %f \n", 
		 iNode,
		 iDof_Node,
		 Node_f_BCVec[iNode][iDof_Node]);
      }
    }
  }
  
  fclose (fp);
}

void WriteFgrFile (char* fileName)
{
  FILE* fp;
  int iFace;
      
  fp = fopen (fileName, "w");
  assert(fp);

  fprintf (fp, "%d\n", Face_nFaces);

  for (iFace = 0; iFace < Face_nFaces; iFace++) {
    fprintf (fp, "%d %d\n", 
	     Face_elementId[iFace], Face_faceId[iFace]);
  }
  
  fclose (fp);
}

void WriteGlobalFile (char* fileName)
{
  FILE* fp;
      
  assert(0 < Node_nGlobalNodes);
  assert(0 < Domain_nDomains);
  assert(Domain_nDomains <= Domain_maxDomains);

  fp = fopen (fileName, "w");
  assert(fp);

  fprintf (fp, "%d\n", Node_nGlobalNodes);
  fprintf (fp, "%d\n", Domain_nDomains);
  
  fclose (fp);
}

int main (int argc, char* argv[])
{
  int iDomain;

  if (argc < 6) {
    fprintf (stderr, 
	     " make_input_file n_domains global_file "
	     " domain_file input_file fgr_file \n");
    exit (1);
  }
  
  Domain_nDomains = atoi (argv[1]);

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[1024];

    sprintf (fileName, "advhddm_in_%d.adv", iDomain);
    ReadAdvIOFile (fileName);

    CheckMesh ();

    sprintf (fileName, "%s.%d", argv[3], iDomain);
    WriteDomainFile (fileName);
    
    sprintf (fileName, "%s.%d", argv[4], iDomain);
    WriteInputFile (fileName);

    sprintf (fileName, "%s.%d", argv[5], iDomain);
    WriteFgrFile (fileName);
  }

  WriteGlobalFile (argv[2]);

  return 0;
}

