/* es_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autolinas_predefined.h>

#include "matrix.h"
#include "es_2d_tri.h"



/* Shape Function */

#define Shape_nDimensions 2
#define Shape_maxOrders 3
#define Shape_maxIPs 7
#define Shape_nNodes Element_nNodes

/* number of integral point */
static int Shape_nIPs[Shape_maxOrders] = {
  1, 3, 7
};

/* local coordinate at each integral point */
static double Shape_L_IPVec
[Shape_maxOrders][Shape_maxIPs][Shape_nDimensions] = {
  { 
    { 1.0/3, 1.0/3 }
  },
  {
    { 1.0/2, 1.0/2 }, 
    { 0.0, 1.0/2 }, 
    { 1.0/2, 0.0 }
  },
  {
    { 1.0/3, 1.0/3 }, 
    { 1.0/2, 1.0/2 },  
    { 0.0, 1.0/2 }, 
    { 1.0/2, 0.0 }, 
    { 1.0, 0.0 },  
    { 0.0, 1.0 }, 
    { 0.0, 0.0 }
  }
};

/* weight value at each integral point for numerical integration */
static double Shape_weight_IP[Shape_maxOrders][Shape_maxIPs] = {
  { 1.0/2 }, 
  { 1.0/6, 1.0/6, 1.0/6 }, 
  { 27.0/120, 8.0/120, 8.0/120, 8.0/120, 3.0/120, 3.0/120, 3.0/120 }
};

/* shape function vector */
static void Shape_NVec 
(double LVec[Shape_nDimensions], 
 double RETURN_VALUE[Shape_nNodes])
{
  /* _{N} = Vector (1.0 - {L}.at (1) - {L}.at (2), 
     {L}.at (1), {L}.at (2)); */
  RETURN_VALUE[0] = 1.0 - LVec[0] - LVec[1];
  RETURN_VALUE[1] = LVec[0];
  RETURN_VALUE[2] = LVec[1];
}

/* derivative of shape function matrix with local coordinate */
static void Shape_dNVec_dLVec 
(double LVec[Shape_nDimensions],
 double RETURN_VALUE[Shape_nDimensions][Shape_nNodes])
{
  /*_d{N}_d{L} = Matrix ((-1.0, 1.0, 0.0), (-1.0, 0.0, 1.0)); */
  RETURN_VALUE[0][0] = -1.0;
  RETURN_VALUE[0][1] = 1.0;
  RETURN_VALUE[0][2] = 0.0;
  RETURN_VALUE[1][0] = -1.0;
  RETURN_VALUE[1][1] = 0.0;
  RETURN_VALUE[1][2] = 1.0;
}



/* Material */

/* 2D plane strain */

#define Material_nStressComponents 3  /* sigma_x, sigma_y, tau_xy */

/* gravity force vector */
double Material_gVec[Node_nDofs];

/* young's modulus */
double Material_E;

/* poisson ratio */
double Material_nyu;

/* mass density */
double Material_rou;

static void Material_DMat
(double RETURN_VALUE
 [Material_nStressComponents][Material_nStressComponents])
{
  double a;
  double c1;
  double c2;

  /* Scalar a = _E*(1 - _nyu)/((1 + _nyu)*(1 - 2*_nyu)); */
  a = Material_E * (1.0 - Material_nyu) 
    / ((1.0 + Material_nyu) * (1.0 - 2.0 * Material_nyu));

  /* Scalar c1 = _nyu/(1 - _nyu); */
  c1 = Material_nyu / (1.0 - Material_nyu);

  /* Scalar c2 = (1 - 2*_nyu) / (2*(1 - _nyu)); */
  c2 = (1.0 - 2.0 * Material_nyu) / (2.0 * (1.0 - Material_nyu));

  /* _[D] = Matrix (
     (a,	a*c1,	0.0), 
     (a*c1,	a,	0.0), 
     (0.0,	0.0,	a*c2)
     ); */
  RETURN_VALUE[0][0] = a;
  RETURN_VALUE[0][1] = a * c1;
  RETURN_VALUE[0][2] = 0.0;
  RETURN_VALUE[1][0] = a * c1;
  RETURN_VALUE[1][1] = a;
  RETURN_VALUE[1][2] = 0.0;
  RETURN_VALUE[2][0] = 0.0;
  RETURN_VALUE[2][1] = 0.0;
  RETURN_VALUE[2][2] = a * c2;
}



/* Node */

/* 2D plane strain, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* displacement boundary condition */
int Node_u_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition */
int Node_f_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Node_maxNodes][Node_nDofs];

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];





/* Element */

/* 2D plane strain, linear elastic */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];

#define Element_orderOfStiffnessMatrix 1
#define Element_orderOfBodyForceVector 1

static void Move_NodeVec_ElementMat
(int iElement,
 double nv[Node_maxNodes][Node_nDofs],
 double em[Element_nNodes][Node_nDofs])
{
  int iNode_Elem;

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = Element_nodeIds[iElement][iNode_Elem];
    int iDof_Node;

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      em[iNode_Elem][iDof_Node] = nv[iNode][iDof_Node];
    }
  }
}

/* jacobian matrix */
static void Element_JMat
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Space_nDimensions][Space_nDimensions])
{
  double XMat[Element_nNodes][Space_nDimensions];
  double tmp0[Shape_nDimensions][Shape_nNodes];
  
  /* Matrix (_nNodes, _space._nDimensions) [X] = _nodeShape._{X}; */
  Move_NodeVec_ElementMat (iElement, Node_XVec, 
			   XMat);
  
  /* _[J] = _normalizedShape._d{N}_d{L} ({L}) * [X]; */
  Shape_dNVec_dLVec (LVec, 
		     tmp0);
  Mul_Mat_Mat_Mat 
    (tmp0, Shape_nDimensions, Shape_nNodes,
     XMat, Element_nNodes, Space_nDimensions, 
     RETURN_VALUE, Space_nDimensions, Space_nDimensions);
}

/* shape function vector */
static void Element_NVec
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Element_nNodes])
{
  /* _{N} = _normalizedShape._{N} ({L}); */
  Shape_NVec (LVec, 
	      RETURN_VALUE);
}

/* derivative of shape function matrix by global coordinate */
static void Element_dNVec_dXVec 
(int iElement,
 double LVec[Space_nDimensions], 
 double RETURN_VALUE[Space_nDimensions][Element_nNodes])
{
  double tmp0[Space_nDimensions][Space_nDimensions];
  double tmp1[Space_nDimensions][Space_nDimensions];
  double tmp2[Shape_nDimensions][Shape_nNodes];

  /* _d{N}_d{X} = _[J] ({L}).inverse 
   * _normalizedShape._d{N}_d{L} ({L});  */
  Element_JMat (iElement, LVec, 
		tmp0);
  Inverse_Mat_Mat (tmp0, 
		   tmp1);
  Shape_dNVec_dLVec (LVec,
		     tmp2);
  Mul_Mat_Mat_Mat 
    (tmp1, Space_nDimensions, Space_nDimensions, 
     tmp2, Shape_nDimensions, Shape_nNodes,
     RETURN_VALUE, Space_nDimensions, Element_nNodes);
}

/* number of integral point */
static void Element_nIPs 
(int iElement,
 int order,
 int* RETURN_VALUE)
{
  /* _nIPs = _normalizedShape._nIPs (order - 1); */
  *RETURN_VALUE = Shape_nIPs[order - 1];
}

/* local coordinate at each integral point */
static void Element_L_IPVec 
(int iElement,
 int order, 
 int iIP, 
 double RETURN_VALUE[Space_nDimensions])
{
  /* _{L_IP} = _normalizedShape._{L_IP} (order - 1, iIP); */
  Move_Vec_Vec (Shape_L_IPVec[order - 1][iIP], Shape_nDimensions, 
		RETURN_VALUE);
}

/* coefficient value (weight * volume) 
   at each integral point for numerical integration */
static void Element_V_IP
(int iElement,
 int order, 
 int iIP,
 double* RETURN_VALUE)
{
  double LVec[Space_nDimensions];
  double tmp0[Space_nDimensions][Space_nDimensions];
  double tmp1;

  /* Vector (_space._nDimensions) {L} 
     = _normalizedShape._{L_IP} (order - 1, iIP); */
  Move_Vec_Vec (Shape_L_IPVec[order - 1][iIP], Shape_nDimensions, 
		LVec);

  /* _V_IP = _[J] ({L}).determinant 
   * _normalizedShape._weight_IP (order - 1, iIP); */
  Element_JMat (iElement, LVec, 
		tmp0);
  Determinant_Mat_Sca (tmp0, 
		       &tmp1);

  *RETURN_VALUE = tmp1 * Shape_weight_IP[order - 1][iIP];
}

/* shape function matrix */
static void Element_NMat 
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Node_nDofs][Element_nDofs])
{
  double NVec[Element_nNodes];
  int iNode;
  int base;
  double value;

  /* Vector (_elementShape._nNodes) {N} 
     = _elementShape._{N} ({L}); */
  Element_NVec (iElement, LVec,
		NVec);

  /* _[N].clear; */
  Clear_Mat 
    (RETURN_VALUE, Node_nDofs, Element_nDofs);

  for (iNode = 0; iNode < Element_nNodes; iNode++) {

    /* Integer base = iNode * _node._nDofs; */
    base = iNode * Node_nDofs;

    /* Scalar value = {N}.at (iNode); */
    value = NVec[iNode];

    /* _[N].at (0, base + 0) = value; */
    RETURN_VALUE[0][base + 0] = value;
    
    /* _[N].at (1, base + 1) = value; */
    RETURN_VALUE[1][base + 1] = value;

  }
}

/* displacement strain matrix */
static void Element_BMat
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Material_nStressComponents][Element_nDofs])
{
  double dNVec_dXVec[Space_nDimensions][Element_nNodes];
  int iNode;
  int base;

  /* Matrix (_space._nDimensions, _elementShape._nNodes) d{N}_d{X} 
     = _elementShape._d{N}_d{X} ({L}); */
  Element_dNVec_dXVec (iElement, LVec,
		       dNVec_dXVec);

  /* _[B].clear; */
  Clear_Mat 
    (RETURN_VALUE, Material_nStressComponents, Element_nDofs);

  for (iNode = 0; iNode < Element_nNodes; iNode++) {

    /* Integer base = iNode * _node._nDofs; */
    base = iNode * Node_nDofs;

    /* _[B].at (0, base + 0) = d{N}_d{X}.at (0, iNode); */
    RETURN_VALUE[0][base + 0] = dNVec_dXVec[0][iNode];

    /* _[B].at (1, base + 1) = d{N}_d{X}.at (1, iNode); */
    RETURN_VALUE[1][base + 1] = dNVec_dXVec[1][iNode];

    /* _[B].at (2, base + 0) = d{N}_d{X}.at (1, iNode); */
    RETURN_VALUE[2][base + 0] = dNVec_dXVec[1][iNode];

    /* _[B].at (2, base + 1) = d{N}_d{X}.at (0, iNode); */
    RETURN_VALUE[2][base + 1] = dNVec_dXVec[0][iNode];

  }
}

/* stiffness matrix */
void Element_KMat
(int iElement,
 double RETURN_VALUE[Element_nDofs][Element_nDofs])
{
  int order;
  int nIPs;
  int iIP;
  double LVec[Space_nDimensions];
  double V;
  double BMat[Material_nStressComponents][Element_nDofs];
  double tmp0[Material_nStressComponents][Material_nStressComponents];
  double tmp1[Element_nDofs][Material_nStressComponents];
  double tmp2[Element_nDofs][Element_nDofs];

  /* Integer order = _orderOfStiffnessMatrix; */
  order = Element_orderOfStiffnessMatrix;

  /* Integer nIPs = _elementShape._nIPs (order); */
  Element_nIPs (iElement, order,
		&nIPs);

  /* _[K].clear; */
  Clear_Mat 
    (RETURN_VALUE, Element_nDofs, Element_nDofs);

  for (iIP = 0; iIP < nIPs; iIP++) {

    /* Vector (_space._nDimensions) {L} 
       = _elementShape._{L_IP} (order, iIP); */
    Element_L_IPVec (iElement, order, iIP,
		     LVec);

    /* Scalar V = _elementShape._V_IP (order, iIP); */
    Element_V_IP (iElement, order, iIP,
		  &V);

    /* Matrix (_material._nStressComponents, _nDofs) [B] 
       = _[B] ({L}); */
    Element_BMat (iElement, LVec,
		  BMat);

    /* _[K] = _[K] + [B].transpose * _material._[D] * [B] * V; */
    Material_DMat 
      (tmp0);
    Mul_MatTr_Mat_Mat
      (BMat, Material_nStressComponents, Element_nDofs,
       tmp0, Material_nStressComponents, Material_nStressComponents,
       tmp1, Element_nDofs, Material_nStressComponents);
    Mul_Mat_Mat_Mat
      (tmp1, Element_nDofs, Material_nStressComponents,
       BMat, Material_nStressComponents, Element_nDofs,
       tmp2, Element_nDofs, Element_nDofs);
    MulAdd_Mat_Sca_Mat
      (tmp2, Element_nDofs, Element_nDofs,
       V,
       RETURN_VALUE);

  }
}

/* body force vector */
void Element_f_bVec
(int iElement,
 double RETURN_VALUE[Element_nDofs])
{
  double bodyForce[Node_nDofs];
  int order;
  int nIPs;
  int iIP;
  double LVec[Space_nDimensions];
  double V;
  double tmp0[Node_nDofs][Element_nDofs];
  double tmp1[Element_nDofs];

  /* Vector (_node._nDofs) {bodyForce} 
     = _material._{g} * _material._rou; */
  Mul_Vec_Sca_Vec
    (Material_gVec, Space_nDimensions,
     Material_rou,
     bodyForce);

  /* Integer order = _orderOfBodyForceVector; */
  order = Element_orderOfBodyForceVector;

  /* Integer nIPs = _elementShape._nIPs (order); */
  Element_nIPs (iElement, order,
		&nIPs);

  /* _{f_b}.clear; */
  Clear_Vec
    (RETURN_VALUE, Element_nDofs);

  for (iIP = 0; iIP < nIPs; iIP++) {

    /* Vector (_space._nDimensions) {L} 
       = _elementShape._{L_IP} (order, iIP); */
    Element_L_IPVec (iElement, order, iIP,
		     LVec);

    /* Scalar V = _elementShape._V_IP (order, iIP); */
    Element_V_IP (iElement, order, iIP,
		  &V);

    /* _{f_b} = _{f_b} + _[N] ({L}).transpose * {bodyForce} * V; */
    Element_NMat (iElement, LVec,
		  tmp0);
    Mul_MatTr_Vec_Vec
      (tmp0, Node_nDofs, Element_nDofs,
       bodyForce, Node_nDofs,
       tmp1, Element_nDofs);
    MulAdd_Vec_Sca_Vec
      (tmp1, Element_nDofs,
       V,
       RETURN_VALUE);

  }
}



