/* matrix.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"

#include "matrix.h"



#define PRINT_EPS 0.000001

/* WARNING : returning static variable */
char* RealString (double value, int digit0, int digit1)
{
  static char format[80];
  static char buf[256];
  int length;

  if (fabs (value) < PRINT_EPS) {
    value = 0.0;
  }
  sprintf (format, " %d.%d", digit0 + 4, digit1 + 4);
  format[0] = '%';
  strcat (format, "lf");

  sprintf (buf, format, value);
  length = strlen (buf);
  buf[length - 4] = 0;
  return buf;
}



/* basic matrix and vector operations */

#define MATRIX_AT(mat, x, y, nColumns) ((double*)mat)[(x)*(nColumns)+(y)]

void Clear_Vec
(double vec[], int vecNDimensions)
{ 
  int i; 

  for (i = 0; i < vecNDimensions; i++) { 
    vec[i] = 0.0; 
  }
}

void Move_Vec_Vec
(double vec0[], int vecNDimensions,
 double vec1[])
{ 
  int i;

  for (i = 0; i < vecNDimensions; i++) { 
    vec1[i] = vec0[i]; 
  }
}

void Mul_Vec_Sca_Vec
(double vec0[], int vecNDimensions,
 double sca,
 double vec1[])
{ 
  int i; 

  for (i = 0; i < vecNDimensions; i++) { 
    vec1[i] = vec0[i] * sca; 
  } 
}

void MulAdd_Vec_Sca_Vec
(double vec0[], int vecNDimensions,
 double sca,
 double vec1[])
{ 
  int i; 

  for (i = 0; i < vecNDimensions; i++) { 
    vec1[i] += vec0[i] * sca; 
  } 
}

void Clear_Mat 
(void* mat, int matNRows, int matNColumns)
{ 
  int i, j; 

  for (i = 0; i < matNRows; i++) {
    for (j = 0; j < matNColumns; j++) { 
      MATRIX_AT(mat, i, j, matNColumns) = 0.0; 
    } 
  } 
}

void Mul_Mat_Vec_Vec
(void* mat, int matNRows, int matNColumns,
 double vec0[], int vec0NDimensions, 
 double vec1[], int vec1NDimensions)
/* mat[matNRows][matNColumns] */
/* vec0[matNRows] */
/* vec1[matNColumns] */
{ 
  int i, j; 

  assert(vec0NDimensions == matNColumns);
  assert(vec1NDimensions == matNRows);

  for (i = 0; i < matNRows; i++) { 
    vec1[i] = 0.0; 
    for (j = 0; j < matNColumns; j++) { 
      vec1[i] += MATRIX_AT(mat, i, j, matNColumns) * vec0[j]; 
    } 
  } 
}

void Mul_MatTr_Vec_Vec
(void* mat, int matNRows, int matNColumns,
 double vec0[], int vec0NDimensions, 
 double vec1[], int vec1NDimensions)
/* mat[matNRows][matNColumns] */
/* vec0[matNRows] */
/* vec1[matNColumns] */
{ 
  int i, j; 

  assert(vec0NDimensions == matNRows);
  assert(vec1NDimensions == matNColumns);

  for (i = 0; i < matNColumns; i++) { 
    vec1[i] = 0.0; 
    for (j = 0; j < matNRows; j++) { 
      vec1[i] += MATRIX_AT(mat, j, i, matNColumns) * vec0[j]; 
    } 
  } 
}

void Mul_Mat_Mat_Mat
(void* mat0, int mat0NRows, int mat0NColumns,
 void* mat1, int mat1NRows, int mat1NColumns, 
 void* mat2, int mat2NRows, int mat2NColumns)
/* mat0[mat0NRows][mat0NColumns] */
/* mat1[mat0NColumns][mat1NColumns] */
/* mat2[mat0NRows][mat1NColumns] */   
{ 
  int i, j, k; 

  assert(mat0NColumns == mat1NRows);
  assert(mat2NRows == mat0NRows);
  assert(mat2NColumns == mat1NColumns);

  for (i = 0; i < mat0NRows; i++) {
    for (j = 0; j < mat1NColumns; j++) { 
      MATRIX_AT(mat2, i, j, mat2NColumns) = 0.0;
      for (k = 0; k < mat0NColumns; k++) { 
	MATRIX_AT(mat2, i, j, mat2NColumns) 
	  += MATRIX_AT(mat0, i, k, mat0NColumns) 
	  * MATRIX_AT(mat1, k, j, mat1NColumns);
      } 
    } 
  } 
}

void Mul_MatTr_Mat_Mat
(void* mat0, int mat0NRows, int mat0NColumns,
 void* mat1, int mat1NRows, int mat1NColumns, 
 void* mat2, int mat2NRows, int mat2NColumns)
/* mat0[mat0NRows][mat0NColumns] */
/* mat1[mat0NRows][mat1NColumns] */
/* mat2[mat0NColumns][mat1NColumns] */
{ 
  int i, j, k; 

  assert(mat1NRows == mat0NRows);
  assert(mat2NRows == mat0NColumns);
  assert(mat2NColumns == mat1NColumns);

  for (i = 0; i < mat0NColumns; i++) {
    for (j = 0; j < mat1NColumns; j++) { 
      MATRIX_AT(mat2, i, j, mat2NColumns) = 0.0;
      for (k = 0; k < mat0NRows; k++) { 
	MATRIX_AT(mat2, i, j, mat2NColumns) 
	  += MATRIX_AT(mat0, k, i, mat0NColumns) 
	  * MATRIX_AT(mat1, k, j, mat1NColumns);
      } 
    } 
  } 
}

void MulAdd_Mat_Sca_Mat
(void* mat0, int matNRows, int matNColumns,
 double sca,
 void* mat1)
{ 
  int i, j; 

  for (i = 0; i < matNRows; i++) {
    for (j = 0; j < matNColumns; j++) { 
      MATRIX_AT(mat1, i, j, matNColumns)
	+= MATRIX_AT(mat0, i, j, matNColumns) * sca;
    } 
  } 
}

void Determinant_Mat_Sca
(double mat[Space_nDimensions][Space_nDimensions],
 double* sca)
{
#if (Space_nDimensions == 2)
  *sca = mat[0][0] * mat[1][1] - mat[0][1] * mat[1][0];
#endif
  
#if (Space_nDimensions == 3)
  *sca = mat[0][0] * mat[1][1] * mat[2][2] 
    + mat[0][1] * mat[1][2] * mat[2][0]
    + mat[0][2] * mat[1][0] * mat[2][1]
    - mat[2][0] * mat[1][1] * mat[0][2]
    - mat[2][1] * mat[1][2] * mat[0][0]
    - mat[2][2] * mat[1][0] * mat[0][1];
#endif
}

void Inverse_Mat_Mat
(double mat0[Space_nDimensions][Space_nDimensions],
 double mat1[Space_nDimensions][Space_nDimensions])
{
  double det;

  Determinant_Mat_Sca (mat0, &det);
#if (Space_nDimensions == 2)
  mat1[0][0] = mat0[1][1] / det;
  mat1[1][1] = mat0[0][0] / det;
  mat1[0][1] = mat0[0][1] * -1.0 / det;
  mat1[1][0] = mat0[1][0] * -1.0 / det;
#endif

#if (Space_nDimensions == 3)
  mat1[0][0] = (mat0[1][1] * mat0[2][2] - mat0[2][1] * mat0[1][2]) / det;
  mat1[0][1] = (mat0[0][2] * mat0[2][1] - mat0[0][1] * mat0[2][2]) / det;
  mat1[0][2] = (mat0[0][1] * mat0[1][2] - mat0[0][2] * mat0[1][1]) / det;
  mat1[1][0] = (mat0[2][0] * mat0[1][2] - mat0[1][0] * mat0[2][2]) / det;
  mat1[1][1] = (mat0[0][0] * mat0[2][2] - mat0[0][2] * mat0[2][0]) / det;
  mat1[1][2] = (mat0[1][0] * mat0[0][2] - mat0[1][2] * mat0[0][0]) / det;
  mat1[2][0] = (mat0[1][0] * mat0[2][1] - mat0[2][0] * mat0[1][1]) / det;
  mat1[2][1] = (mat0[0][1] * mat0[2][0] - mat0[2][1] * mat0[0][0]) / det;
  mat1[2][2] = (mat0[0][0] * mat0[1][1] - mat0[0][1] * mat0[1][0]) / det;
#endif
}



