/* make_input_file.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <Adv/AdvDocument.h>

#include "../lib/autolinas_predefined.h"



int Node_nNodes;
int Element_nElements;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];

/* displacement boundary condition */
int Node_u_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition */
int Node_f_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Node_maxNodes][Node_nDofs];



void ReadNode (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nDimensions;

  doc = adv_dbox_find_by_property 
    (dbox, NULL, "content_type", "Node", NULL);
  assert(doc != NULL);

  adv_dio_get_property_int32 (doc, "num_items", 
			      &Node_nNodes);
  assert(0 < Node_nNodes);
  assert(Node_nNodes <= Node_maxNodes);

  fprintf (stderr, " Node_nNodes %d \n", Node_nNodes);

  adv_dio_get_property_int32 (doc, "dimension", &nDimensions);
  assert(nDimensions == 3);

  offset = 0;

  offset += adv_dio_read_float64v (doc, offset, Node_nNodes * nDimensions, 
				   (void*)Node_XVec);

  adv_dio_close (doc);
}

void ReadElement (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  char elementTypeName[256];
  int nNodesPerElement;
  int iElement;
  int iNode_Elem;

  doc = adv_dbox_find_by_property 
    (dbox, NULL, "content_type", "Element", NULL);
  assert(doc != NULL);

  strcpy (elementTypeName, adv_dio_get_property (doc, "element_type"));
  if (strcmp (elementTypeName, "3DLinearTetrahedron") == 0) {
    assert(Element_nNodes == 4);
  } else if (strcmp (elementTypeName, "3DQuadraticTetrahedron") == 0) {
    assert(Element_nNodes == 10);
  } else {
    assert(0);
  }

  adv_dio_get_property_int32 (doc, "num_items", 
			      &Element_nElements);
  assert(0 < Element_nElements);
  assert(Element_nElements <= Element_maxElements);

  fprintf (stderr, " Element_nElements %d \n", Element_nElements);

  adv_dio_get_property_int32 (doc, "num_nodes_per_element", 
			      &nNodesPerElement);
  assert(nNodesPerElement == Element_nNodes);

  offset = 0;

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      offset += adv_dio_read_int32 (doc, offset, 
				    &Element_nodeIds[iElement][iNode_Elem]);
    }
  }
      
  adv_dio_close (doc);
}

void ReadDisplacementBc (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nItems;
  int iItem;
  int iNode;
  int iDof_Node;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "FEGenericAttribute", 
     "label", "ForcedDisplacement", NULL);
  if (doc == NULL) {
    return;
  }

  offset = 0;

  adv_dio_get_property_int32 (doc, "num_items", 
			      &nItems);
  assert(0 <= nItems);

  for (iItem = 0; iItem < nItems; iItem++) {
    int nodeId;
    int dofId;
    double value;

    offset += adv_dio_read_int32 (doc, offset, 
				  &nodeId);
    offset += adv_dio_read_int32 (doc, offset, 
				  &dofId);
    offset += adv_dio_read_float64 (doc, offset, 
				    &value);

    Node_u_BCVecFlag[nodeId][dofId] = 1;
    Node_u_BCVec[nodeId][dofId] = value;
  }

  adv_dio_close (doc);
}

void ReadLoadBc (AdvDatabox* dbox)
{
  AdvDocument* doc;
  adv_off_t offset;
  int nItems;
  int iItem;
  int iNode;
  int iDof_Node;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_f_BCVecFlag[iNode][iDof_Node] = 0;
      Node_f_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  doc = adv_dbox_find_by_property
    (dbox, NULL, 
     "content_type", "FEGenericAttribute", 
     "label", "Load", NULL);
  if (doc == NULL) {
    return;
  }

  offset = 0;

  adv_dio_get_property_int32 (doc, "num_items", 
			      &nItems);
  assert(0 <= nItems);

  for (iItem = 0; iItem < nItems; iItem++) {
    int nodeId;
    int dofId;
    double value;

    offset += adv_dio_read_int32 (doc, offset, 
				  &nodeId);
    offset += adv_dio_read_int32 (doc, offset, 
				  &dofId);
    offset += adv_dio_read_float64 (doc, offset, 
				    &value);

    Node_f_BCVecFlag[nodeId][dofId] = 1;
    Node_f_BCVec[nodeId][dofId] = value;
  }

  adv_dio_close (doc);
}

void ReadAdvIOFile (char* fileName)
{
  AdvDatabox* dbox;

  dbox = adv_dbox_new ();

  if (!adv_dbox_add (dbox, fileName)) {
    fprintf (stderr, "Cannot open %s\n", fileName);
    exit (1);
  }

  ReadNode (dbox);
  ReadElement (dbox);
  ReadDisplacementBc (dbox);
  ReadLoadBc (dbox);

  adv_dbox_close (dbox);
}

void CheckMesh ()
{
  static int nodeFlag[Node_maxNodes];
  int iNode;
  int iElement;
  int iNode_Elem;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    nodeFlag[iNode] = 0;
  }

  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int nodeId = Element_nodeIds[iElement][iNode_Elem];

      assert(0 <= nodeId);
      assert(nodeId < Node_nNodes);
      nodeFlag[nodeId] = 1;
    }
  }

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    assert(nodeFlag[iNode]);
  }
}

void WriteInputFile (char* fileName)
{
  FILE* fp;
  int iElement;
  int iNode_Elem;
  int iNode;
  int iDof_Node;
  int count;
      
  fp = fopen (fileName, "w");
  assert(fp);

  /* element connectivity */

  fprintf (fp, "%d\n", Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iNode_Elem]);
    }

    fprintf (fp, "\n");
  }
  
  /* nodal coordinate */

  fprintf (fp, "%d\n", Node_nNodes);

  /* ordered by part node ID */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%f %f %f \n", 
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1],
	     Node_XVec[iNode][2]);
  }
  
  /* displacement B.C. */

  count = 0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) count++;
    }
  }

  fprintf (fp, "%d\n", count);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) {
	fprintf (fp, "%d %d %f \n", 
		 iNode,
		 iDof_Node,
		 Node_u_BCVec[iNode][iDof_Node]);
      }
    }
  }

  /* load B.C. */

  count = 0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) count++;
    }
  }

  fprintf (fp, "%d\n", count);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) {
	fprintf (fp, "%d %d %f \n", 
		 iNode,
		 iDof_Node,
		 Node_f_BCVec[iNode][iDof_Node]);
      }
    }
  }
  
  fclose (fp);
}

int main (int argc, char* argv[])
{
  if (argc < 3) {
    fprintf (stderr, 
	     " make_input_file advio_file input_file \n");
    exit (1);
  }
  
  ReadAdvIOFile (argv[1]);
  
  CheckMesh ();
  
  WriteInputFile (argv[2]);

  return 0;
}

