/* post_es_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"

#define Element_maxElements 1000000



/* Node */

/* 2D plane strain, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];

/* stress analysis result */
double Node_sigma_eq[Node_maxNodes];



/* Element */

/* 2D plane strain, linear elastic */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &Node_nNodes);
  assert(0 < Node_nNodes);
  assert(Node_nNodes <= Node_maxNodes);

  fprintf (stderr, " nNodes %d \n", Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId,
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iNode_Elem;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &Element_nElements);
  assert(0 < Element_nElements);
  assert(Element_nElements <= Element_maxElements);

  fprintf (stderr, " nElements %d \n", Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId);
    assert(elementId == iElement);
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fclose (fp);
}

void ReadDisplacementFile (const char *fileName)
{
  FILE *fp;
  int nNodes;
  int iNode;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nNodes);
  assert(nNodes == Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId,
	    &Node_uVec[iNode][0], 
	    &Node_uVec[iNode][1]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}

void ReadStressFile (const char *fileName)
{
  FILE *fp;
  int nNodes;
  int iNode;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nNodes);
  assert(nNodes == Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &Node_sigma_eq[iNode]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}


double MinRange = 0.0;
double MaxRange = 1.0;
int NGrades = 20;

double Scale = 2.0;

int MeshIsShown = 1;
int DeformationIsShown;
int ContourIsShown;

#define CONTOUR_SMOOTH 0
#define CONTOUR_LINE 1

int ContourType = CONTOUR_SMOOTH;



static void SetRange ()
{
  int iNode;
  double minX, minY;
  double maxX, maxY;
  double size;
  double midX, midY;
  double maxLength;

  minX = Node_XVec[0][0];
  minY = Node_XVec[0][1];

  maxX = Node_XVec[0][0];
  maxY = Node_XVec[0][1];
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double x = Node_XVec[iNode][0];
    double y = Node_XVec[iNode][1];
    
    AutoGL_GetMinVector2D (&minX, &minY,
			   minX, minY,
			   x, y);
    AutoGL_GetMaxVector2D (&maxX, &maxY,
			   maxX, maxY,
			   x, y);
  }

  size = AutoGL_GetVectorDistance2D 
    (minX, minY,
     maxX, maxY);
  midX = (minX + maxX) / 2;
  midY = (minY + maxY) / 2;

  AutoGL_SetViewSize (size / 2); 
  AutoGL_SetViewCenter (midX, midY, 0.0); 

  MinRange = Node_sigma_eq[0]; 
  MaxRange = Node_sigma_eq[0]; 
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_sigma_eq[iNode] < MinRange) {
      MinRange = Node_sigma_eq[iNode];
    }
    if (MaxRange < Node_sigma_eq[iNode]) {
      MaxRange = Node_sigma_eq[iNode];
    }
  }

  maxLength = 0.0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double ux = Node_uVec[iNode][0];
    double uy = Node_uVec[iNode][1];
    double length = AutoGL_GetVectorLength2D (ux, uy);

    if (maxLength < length) {
      maxLength = length;
    }
  }

  Scale = size / maxLength * 0.1;
}

void PlotMesh (void) 
{
  int iElement;

  AutoGL_SetColor (1, 1, 1);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];

    AutoGL_DrawLine (Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0,
		     Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0);
    AutoGL_DrawLine (Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0,
		     Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0);
    AutoGL_DrawLine (Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0,
		     Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0);
  }
}

void PlotSmoothContour (void) 
{
  int iElement;

  AutoGL_DrawContourMap (80, MinRange, MaxRange);

  AutoGL_SetColor (1, 1, 1);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];

    AutoGL_DrawSmoothContourTriangle 
      (MinRange, MaxRange,
       Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0, 
       Node_sigma_eq[nodeId0],
       Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0, 
       Node_sigma_eq[nodeId1],
       Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0, 
       Node_sigma_eq[nodeId2]);
  }
}

void PlotLineContour (void) 
{
  int iElement;

  AutoGL_DrawLineContourMap (80, MinRange, MaxRange, NGrades);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];
    
    AutoGL_DrawLineContourTriangle 
      (MinRange, MaxRange, NGrades,
       Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0, 
       Node_sigma_eq[nodeId0],
       Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0, 
       Node_sigma_eq[nodeId1],
       Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0, 
       Node_sigma_eq[nodeId2]);
  }
}

void PlotDeformation (void) 
{
  int iElement;

  AutoGL_SetColor (0, 1, 1);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];

    double p0X = Node_XVec[nodeId0][0] + Node_uVec[nodeId0][0] * Scale;
    double p1X = Node_XVec[nodeId1][0] + Node_uVec[nodeId1][0] * Scale;
    double p2X = Node_XVec[nodeId2][0] + Node_uVec[nodeId2][0] * Scale;

    double p0Y = Node_XVec[nodeId0][1] + Node_uVec[nodeId0][1] * Scale;
    double p1Y = Node_XVec[nodeId1][1] + Node_uVec[nodeId1][1] * Scale;
    double p2Y = Node_XVec[nodeId2][1] + Node_uVec[nodeId2][1] * Scale;

    AutoGL_DrawLine (p0X, p0Y, 0,
		     p1X, p1Y, 0);
    AutoGL_DrawLine (p1X, p1Y, 0,
		     p2X, p2Y, 0);
    AutoGL_DrawLine (p2X, p2Y, 0,
		     p0X, p0Y, 0);
  }
}

void RedrawView (void) 
{
  if (MeshIsShown) {
    PlotMesh ();
  }

  if (DeformationIsShown) {
    PlotDeformation ();
  }
  
  if (ContourIsShown) {
    switch (ContourType) {
    case CONTOUR_SMOOTH:
      PlotSmoothContour ();
      break;
    case CONTOUR_LINE:
      PlotLineContour ();
      break;
    default:
      assert(0);
      break;
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  assert(Space_nDimensions == 2);
  assert(Element_nNodes == 3
	 || Element_nNodes == 6);

  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " post_es_2d_tri  node_file  element_file \n"
       "        displacement_file \n");
    exit (1);
  }
  
  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadDisplacementFile (argv[3]);

  SetRange ();

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode2D (NULL);
  AutoGL_EnableDragInMode2D ();

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode2D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_SetLabel ("mesh");
  AutoGL_AddBoolean (&DeformationIsShown, "DeformationIsShown");
  AutoGL_SetLabel ("deformation");
  AutoGL_AddBoolean (&ContourIsShown, "ContourIsShown");
  AutoGL_SetLabel ("contour");
  AutoGL_AddInteger (&ContourType, "ContourType");
  AutoGL_AddIntegerItem ("smooth"); 
  AutoGL_AddIntegerItem ("line"); 
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddInteger (&NGrades, "NGrades");
  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddCallback (UpdateModel, "UpdateModel");
}






