/* post_es_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"

#define Element_maxElements 1000000



/* Node */

/* 3D solid, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];

/* equivalent stress analysis result */
double Node_sigma_eq[Node_maxNodes];



/* Element */

/* 3D solid, linear elastic */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &Node_nNodes);
  assert(0 < Node_nNodes);
  assert(Node_nNodes <= Node_maxNodes);

  fprintf (stderr, " nNodes %d \n", Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId,
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1], 
	    &Node_XVec[iNode][2]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iNode_Elem;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &Element_nElements);
  assert(0 < Element_nElements);
  assert(Element_nElements <= Element_maxElements);

  fprintf (stderr, " nElements %d \n", Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId);
    assert(elementId == iElement);
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fclose (fp);
}

void ReadDisplacementFile (const char *fileName)
{
  FILE *fp;
  int nNodes;
  int iNode;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nNodes);
  assert(nNodes == Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId,
	    &Node_uVec[iNode][0], 
	    &Node_uVec[iNode][1], 
	    &Node_uVec[iNode][2]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}

void ReadStressFile (const char *fileName)
{
  FILE *fp;
  int nNodes;
  int iNode;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nNodes);
  assert(nNodes == Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &Node_sigma_eq[iNode]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}



double MinRange = 0.0;
double MaxRange = 1.0;

double Scale = 2.0;

int MeshIsShown = 1;
int DeformationIsShown;
int ContourIsShown;



static void SetRange ()
{
  int iNode;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double size;
  double midX, midY, midZ;
  double maxLength;

  minX = Node_XVec[0][0];
  minY = Node_XVec[0][1];
  minZ = Node_XVec[0][2];

  maxX = Node_XVec[0][0];
  maxY = Node_XVec[0][1];
  maxZ = Node_XVec[0][2];

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double x = Node_XVec[iNode][0];
    double y = Node_XVec[iNode][1];
    double z = Node_XVec[iNode][2];

    AutoGL_GetMinVector3D (&minX, &minY, &minZ,
			   minX, minY, minZ,
			   x, y, z);
    AutoGL_GetMaxVector3D (&maxX, &maxY, &maxZ,
			   maxX, maxY, maxZ,
			   x, y, z);
  }

  size = AutoGL_GetVectorDistance3D 
    (minX, minY, minZ,
     maxX, maxY, maxZ);
  midX = (minX + maxX) / 2;
  midY = (minY + maxY) / 2;
  midZ = (minZ + maxZ) / 2;

  AutoGL_SetViewSize (size / 2); 
  AutoGL_SetViewCenter (midX, midY, midZ); 

  MinRange = Node_sigma_eq[0]; 
  MaxRange = Node_sigma_eq[0]; 
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_sigma_eq[iNode] < MinRange) {
      MinRange = Node_sigma_eq[iNode];
    }
    if (MaxRange < Node_sigma_eq[iNode]) {
      MaxRange = Node_sigma_eq[iNode];
    }
  }

  maxLength = 0.0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    double ux = Node_uVec[iNode][0];
    double uy = Node_uVec[iNode][1];
    double uz = Node_uVec[iNode][2]; 
    double length = AutoGL_GetVectorLength3D (ux, uy, uz);

    if (maxLength < length) {
      maxLength = length;
    }
  }

  Scale = size / maxLength * 0.1;
}

void RedrawView (void) 
{
  int iElement;

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];
    int nodeId3 = Element_nodeIds[iElement][3];

    double p0X = Node_XVec[nodeId0][0];
    double p0Y = Node_XVec[nodeId0][1];
    double p0Z = Node_XVec[nodeId0][2];

    double p1X = Node_XVec[nodeId1][0];
    double p1Y = Node_XVec[nodeId1][1];
    double p1Z = Node_XVec[nodeId1][2];

    double p2X = Node_XVec[nodeId2][0];
    double p2Y = Node_XVec[nodeId2][1];
    double p2Z = Node_XVec[nodeId2][2];

    double p3X = Node_XVec[nodeId3][0];
    double p3Y = Node_XVec[nodeId3][1];
    double p3Z = Node_XVec[nodeId3][2];

    if (DeformationIsShown) {
      p0X += Node_uVec[nodeId0][0] * Scale;
      p0Y += Node_uVec[nodeId0][1] * Scale;
      p0Z += Node_uVec[nodeId0][2] * Scale;
      
      p1X += Node_uVec[nodeId1][0] * Scale;
      p1Y += Node_uVec[nodeId1][1] * Scale;
      p1Z += Node_uVec[nodeId1][2] * Scale;
      
      p2X += Node_uVec[nodeId2][0] * Scale;
      p2Y += Node_uVec[nodeId2][1] * Scale;
      p2Z += Node_uVec[nodeId2][2] * Scale;
      
      p3X += Node_uVec[nodeId3][0] * Scale;
      p3Y += Node_uVec[nodeId3][1] * Scale;
      p3Z += Node_uVec[nodeId3][2] * Scale;
    }

    if (MeshIsShown) {
      AutoGL_SetColor (1, 0, 0);
      AutoGL_DrawLine (p0X, p0Y, p0Z,
		       p1X, p1Y, p1Z);
      AutoGL_DrawLine (p1X, p1Y, p1Z,
		       p2X, p2Y, p2Z);
      AutoGL_DrawLine (p2X, p2Y, p2Z,
		       p0X, p0Y, p0Z);
      AutoGL_DrawLine (p0X, p0Y, p0Z,
		       p3X, p3Y, p3Z);
      AutoGL_DrawLine (p1X, p1Y, p1Z,
		       p3X, p3Y, p3Z);
      AutoGL_DrawLine (p2X, p2Y, p2Z,
		       p3X, p3Y, p3Z);
    }

    if (ContourIsShown) {
      AutoGL_DrawSmoothContourTriangle 
	(MinRange, MaxRange,
	 p0X, p0Y, p0Z, Node_sigma_eq[nodeId0], 
	 p1X, p1Y, p1Z, Node_sigma_eq[nodeId1], 
	 p2X, p2Y, p2Z, Node_sigma_eq[nodeId2]);
      AutoGL_DrawSmoothContourTriangle 
	(MinRange, MaxRange,
	 p0X, p0Y, p0Z, Node_sigma_eq[nodeId0],
	 p1X, p1Y, p1Z, Node_sigma_eq[nodeId1],
	 p3X, p3Y, p3Z, Node_sigma_eq[nodeId3]);
      AutoGL_DrawSmoothContourTriangle 
	(MinRange, MaxRange,
	 p1X, p1Y, p1Z, Node_sigma_eq[nodeId1],
	 p2X, p2Y, p2Z, Node_sigma_eq[nodeId2],
	 p3X, p3Y, p3Z, Node_sigma_eq[nodeId3]);
      AutoGL_DrawSmoothContourTriangle 
	(MinRange, MaxRange,
	 p2X, p2Y, p2Z, Node_sigma_eq[nodeId2],
	 p0X, p0Y, p0Z, Node_sigma_eq[nodeId0],
	 p3X, p3Y, p3Z, Node_sigma_eq[nodeId3]);
    } else {
      AutoGL_SetColor (0, 1, 1);
      AutoGL_DrawTriangle (p0X, p0Y, p0Z,
			   p1X, p1Y, p1Z,
			   p2X, p2Y, p2Z);
      AutoGL_DrawTriangle (p0X, p0Y, p0Z,
			   p1X, p1Y, p1Z,
			   p3X, p3Y, p3Z);
      AutoGL_DrawTriangle (p1X, p1Y, p1Z,
			   p2X, p2Y, p2Z,
			   p3X, p3Y, p3Z);
      AutoGL_DrawTriangle (p2X, p2Y, p2Z,
			   p0X, p0Y, p0Z,
			   p3X, p3Y, p3Z);
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  assert(Space_nDimensions == 2);
  assert(Element_nNodes == 4
	 || Element_nNodes == 10);

  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " post_es_3d_tetra  node_file  element_file \n"
       "        displacement_file \n");
    exit (1);
  }
  
  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadDisplacementFile (argv[3]);

  SetRange ();

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_SetLabel ("mesh");
  AutoGL_AddBoolean (&DeformationIsShown, "DeformationIsShown");
  AutoGL_SetLabel ("deformation");
  AutoGL_AddBoolean (&ContourIsShown, "ContourIsShown");
  AutoGL_SetLabel ("contour");
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddCallback (UpdateModel, "UpdateModel");
}






