/* main_es_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autolinas_predefined.h>
#include <autolinas.h>

#include "matrix.h"
#include "es_2d_tri.h"



/* System */

/* 2D plane strain, linear elastic, static analysis */

static void Solve (void)
{
  int iNode;
  int iElement;

  AutoLinAS_Initialize (Node_nNodes);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int iNode_Elem, jNode_Elem;
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int iNode = Element_nodeIds[iElement][iNode_Elem];
      
      for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
	int jNode = Element_nodeIds[iElement][jNode_Elem];
	
	AutoLinAS_RegisterNodalIndex (iNode, jNode);
      }
    }
  }

  AutoLinAS_AllocateMatrix ();

  /* _{u_fixed} = {0}; */
  AutoLinAS_ClearConstraint ();

  /* _{u_fixed} = _node._{u_BC}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int iDof_Node;
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_u_BCVecFlag[iNode][iDof_Node]) {
	double value = Node_u_BCVec[iNode][iDof_Node];

	AutoLinAS_SetConstraint (iNode, iDof_Node, value);
      }
    }
  }

  /* _{f} = {0}; */
  AutoLinAS_ClearRightHandSide ();

  /* _{f} += _element._{f_b}; */
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    double tmp0[Element_nDofs];
    int iNode_Elem;
    
    Element_f_bVec (iElement,
		    tmp0);
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int iNode = Element_nodeIds[iElement][iNode_Elem];
      int iDof_Node;
      
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	int iDof_Elem = iNode_Elem * Node_nDofs + iDof_Node;
	
	AutoLinAS_AddRightHandSide
	  (iNode, iDof_Node, tmp0[iDof_Elem]);
      }
    }
  }

  /* _{f} += _node._{f_BC}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int iDof_Node;
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      if (Node_f_BCVecFlag[iNode][iDof_Node]) {
	double value = Node_f_BCVec[iNode][iDof_Node];

	AutoLinAS_AddRightHandSide (iNode, iDof_Node, value);
      }
    }
  }

  /* _[K] = _element._[K]; */
  /* _[K].fix (_{u_fixed}); */

  AutoLinAS_ClearMatrix ();
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    double em[Element_nDofs][Element_nDofs];
    double nm[Node_nDofs][Node_nDofs];
    int iNode_Elem, jNode_Elem;
    int iDof_Node, jDof_Node;
    
    Element_KMat (iElement,
		  em);
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int iNode = Element_nodeIds[iElement][iNode_Elem];
      
      for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
	int jNode = Element_nodeIds[iElement][jNode_Elem];
	
	for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	  int iDof_Elem = iNode_Elem * Node_nDofs + iDof_Node;
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    int jDof_Elem = jNode_Elem * Node_nDofs + jDof_Node;
	    
	    nm[iDof_Node][jDof_Node] = em[iDof_Elem][jDof_Elem];
	  }
	}
	
	AutoLinAS_AssembleNodalMatrix 
	  (iNode, jNode, nm);
      }
    }
  }
  AutoLinAS_Fix ();

  /* _{u} = _[K].solveSystem (_{f}); */
  AutoLinAS_Solve ();
  
  /* _node._{u} = _{u}; */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int iDof_Node;

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_uVec[iNode][iDof_Node] = AutoLinAS_Result (iNode, iDof_Node);
    }
  }
}

static void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &Node_nNodes);
  assert(0 < Node_nNodes);
  assert(Node_nNodes <= Node_maxNodes);

  fprintf (stderr, " nNodes %d \n", Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    int nodeId;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId,
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1]);
    assert(nodeId == iNode);
  }

  fclose (fp);
}

static void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iNode_Elem;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &Element_nElements);
  assert(0 < Element_nElements);
  assert(Element_nElements <= Element_maxElements);

  fprintf (stderr, " nElements %d \n", Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId);
    assert(elementId == iElement);
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fclose (fp);
}

static void ReadDisplacementNodeDofBcFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  int iDof_Node;
  int nItems;
  int iItem;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &nItems);
  fprintf (stderr, " displacement BCs %d \n", nItems);
  
  for (iItem = 0; iItem < nItems; iItem++) {
    int itemId;
    int nodeId;
    int dofId;
    double value;
    
    fscanf (fp, "%d %d %d %lf", 
	    &itemId, 
	    &nodeId, &dofId, &value);
    assert(itemId == iItem);
    assert(0 <= nodeId);
    assert(nodeId < Node_nNodes);
    assert(0 <= dofId);
    assert(dofId < Node_nDofs);

    Node_u_BCVecFlag[nodeId][dofId] = 1;
    Node_u_BCVec[nodeId][dofId] = value;
  }

  fclose (fp);
}

static void ReadLoadNodeDofBcFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  int iDof_Node;
  int nItems;
  int iItem;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_f_BCVecFlag[iNode][iDof_Node] = 0;
      Node_f_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &nItems);
  fprintf (stderr, " load BCs %d \n", nItems);
  
  for (iItem = 0; iItem < nItems; iItem++) {
    int itemId;
    int nodeId;
    int dofId;
    double value;
    
    fscanf (fp, "%d %d %d %lf", 
	    &itemId, 
	    &nodeId, &dofId, &value);
    assert(itemId == iItem);
    assert(0 <= nodeId);
    assert(nodeId < Node_nNodes);
    assert(0 <= dofId);
    assert(dofId < Node_nDofs);

    Node_f_BCVecFlag[nodeId][dofId] = 1;
    Node_f_BCVec[nodeId][dofId] = value;
  }

  fclose (fp);
}

#define EPS 0.00001

static void MakeBoundaryCondition (void)
{
  int iNode;
  int iDof_Node;
  double minX, maxX;
  double minY, maxY;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iNode][iDof_Node] = 0;
      Node_u_BCVec[iNode][iDof_Node] = 0.0;
    }
  }

  minX = Node_XVec[0][0];
  maxX = Node_XVec[0][0];
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_XVec[iNode][0] < minX) {
      minX = Node_XVec[iNode][0];
    }
    if (maxX < Node_XVec[iNode][0]) {
      maxX = Node_XVec[iNode][0];
    }
  }

  minY = Node_XVec[0][1];
  maxY = Node_XVec[0][1];
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (Node_XVec[iNode][1] < minY) {
      minY = Node_XVec[iNode][1];
    }
    if (maxY < Node_XVec[iNode][1]) {
      maxY = Node_XVec[iNode][1];
    }
  }

#if 0
  /* X axis direction */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (fabs (Node_XVec[iNode][0] - minX) < EPS) {
      /* left edge, constraint */

      Node_u_BCVecFlag[iNode][0] = 1; 
      Node_u_BCVecFlag[iNode][1] = 1; 
      Node_u_BCVec[iNode][0] = 0.0;
      Node_u_BCVec[iNode][1] = 0.0;

    } else if (fabs (Node_XVec[iNode][0] - maxX) < EPS) {
      /* right edge, tension */

      Node_u_BCVecFlag[iNode][0] = 1; 
      Node_u_BCVec[iNode][0] = 0.1;

    }
  }
#endif

#if 1
  /* Y axis direction */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    if (fabs (Node_XVec[iNode][1] - minY) < EPS) {
      /* bottom edge, constraint */

      Node_u_BCVecFlag[iNode][0] = 1; 
      Node_u_BCVecFlag[iNode][1] = 1; 
      Node_u_BCVec[iNode][0] = 0.0;
      Node_u_BCVec[iNode][1] = 0.0;

    } else if (fabs (Node_XVec[iNode][1] - maxY) < EPS) {
      /* top edge, tension */

      Node_u_BCVecFlag[iNode][1] = 1; 
      Node_u_BCVec[iNode][1] = 0.1;

    }
  }
#endif
}

static void MakeOtherCondition (void)
{
  Material_gVec[0] = 0.0;
  Material_gVec[1] = 0.0;

  Material_E = 1.0;
  Material_nyu = 0.3;
  Material_rou = 1.0;
}

static void WriteDisplacementFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Node_nNodes);

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_uVec[iNode][0], 
	     Node_uVec[iNode][1]);
  }

  fclose (fp);
}

int main (int argc, char* argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " main_es_2d_tri  node_file  element_file \n"
       "        displacement_node_dof_bc_file \n"
       "        load_node_dof_bc_file \n"
       "        displacement_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

#if 1
  ReadDisplacementNodeDofBcFile (argv[3]);
  ReadLoadNodeDofBcFile (argv[4]);
#else
  MakeBoundaryCondition ();
#endif
  MakeOtherCondition ();
  
  Solve ();
  
  WriteDisplacementFile (argv[5]);

  return 0;
}


